/*
	ASPI module
	Copyright 1995,1996 (C) ITO Takayuki
*/

#include	<stdio.h>
#include	<io.h>
#include	<fcntl.h>
#ifdef __TURBOC__
#include	<mem.h>
#else
#include	<memory.h>
#include	<dos.h>
#endif
#include	"typedef.h"
#include	"aspiscsi.h"
#include	"misclib.h"

BYTE	ASPICommandStatus = ASPICS_COMPLETE;
BYTE	ASPIAdapterStatus = ASPIAS_NOERROR;
BYTE	SCSIStatus = S_GOOD;
BYTE	SCSISenseKey = 0;
BYTE	SCSISenseCode = 0;
BYTE	SCSISenseCodeQ = 0;
BOOL	UnitAtnRetry = FALSE;
int		Adapters = 0;
WORD	blocksize = 512U;
#ifdef LSI_C
static void	(far *ASPIEntry) (void far *srb, ...);
#else
static void	(far *ASPIEntry) (void far *srb);
#endif

int		InitASPI	(void)

{
	int		fd;
	
#ifdef __TURBOC__
	if ((fd = _open ("SCSIMGR$", O_RDONLY)) == FAILURE)
		return FAILURE;
	if (ioctl (fd, 0x02, &ASPIEntry, sizeof ASPIEntry) == FAILURE)
		return FAILURE;
	_close (fd);
	return SUCCESS;
#else
	union REGS		regs;
	struct SREGS	sregs;
	if (_dos_open ("SCSIMGR$", O_RDONLY, &fd) != SUCCESS)
		return FAILURE;
	regs.x.ax = 0x4402;
	regs.x.bx = fd;
	regs.x.cx = sizeof ASPIEntry;
	regs.x.dx = FP_OFF (&ASPIEntry);
	sregs.ds  = FP_SEG (&ASPIEntry);
	intdosx (&regs, &regs, &sregs);
	if (regs.x.cflag)
		return FAILURE;
	_dos_close (fd);
	return SUCCESS;
#endif
}

char	far *ASPIAdapterInquiry	(int AdapterNo)

{
	static struct InquirySRB	srb;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_ADAPTER_INQUIRY;
	srb.AdapterNo = AdapterNo;
	ASPIEntry (&srb);
	if (srb.Status == ASPICS_COMPLETE)
	{
		Adapters = srb.Adapters;
		return srb.ASPIManagerID;
	}
	else /* ASPICS_INVALID_ADAPTER */
		return NULL;
}

int		ASPIGetDeviceType		(int AdapterNo, int scsiid, int lun)

{
	struct DeviceTypeSRB	srb;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_GET_DEVICE_TYPE;
	srb.AdapterNo = AdapterNo;
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	ASPIEntry (&srb);
	if (srb.Status == ASPICS_COMPLETE)
		return srb.DeviceType;
	else /* ASPICS_NOT_INSTALLED */
		return FAILURE;
}

static int	near ASPIPerform	(struct ExecIOSRB far *psrb)

{
	BYTE	far *sense;
	
	ASPIEntry (psrb);
	while (psrb->Status == ASPICS_PROGRESS);
	ASPIAdapterStatus = psrb->AdapterStatus;
	ASPICommandStatus = psrb->Status;
	SCSIStatus = psrb->TargetStatus;
	if (ASPIAdapterStatus != ASPIAS_NOERROR)
		return FAILURE;
	if (ASPICommandStatus != ASPICS_COMPLETE &&
	    ASPICommandStatus != ASPICS_ERROR)
		return FAILURE;
	if (SCSIStatus == S_GOOD)
		return SUCCESS;
	else if (SCSIStatus == S_CHECK)
	{
		sense = &psrb->cdb [psrb->CDBLength];
		SCSISenseKey   = sense [ 2] & 0x0F;
		SCSISenseCode  = sense [12];
		SCSISenseCodeQ = sense [13];
		if (SCSISenseKey == 0x06 && UnitAtnRetry)
			return ASPIPerform (psrb);
	}
	return FAILURE;
}

int		FormatUnit	(int AdapterNo, int scsiid, int lun)

{
	struct ExecIOSRB	srb;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = RF_DIR_NOXFER;
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 6;
	srb.cdb [0] = 0x04;
#ifdef SCSI_1
	srb.cdb [1] = lun << 5;
#endif
	return ASPIPerform (&srb);
}

int		FormatUnit2	(int AdapterNo, int scsiid, int lun, int flags, 
					int interleave, BYTE far *buf, int buflen)

{
	struct ExecIOSRB	srb;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = (flags & 0x10 ? RF_DIR_TOTARGET : RF_DIR_NOXFER);
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	if (flags & 0x10)
	{
		srb.DataAllocLength = buflen;
		srb.DataBufPtr = buf;
	}
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 6;
	srb.cdb [0] = 0x04;
#ifdef SCSI_1
	srb.cdb [1] = (flags & 0x1F) | (lun << 5);
#else
	srb.cdb [1] =  flags & 0x1F;
#endif
	srb.cdb [3] = interleave >> 8;
	srb.cdb [4] = interleave;
	return ASPIPerform (&srb);
}

int		Inquiry	(int AdapterNo, int scsiid, int lun, BYTE far *buf, int buflen)

{
	struct ExecIOSRB	srb;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = RF_DIR_AUTO;
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	srb.DataAllocLength = buflen;
	srb.DataBufPtr = buf;
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 6;
	srb.cdb [0] = 0x12;
#ifdef SCSI_1
	srb.cdb [1] = lun << 5;
#endif
	srb.cdb [4] = buflen;
	return ASPIPerform (&srb);
}

int		TestUnitReady	(int AdapterNo, int scsiid, int lun)

{
	struct ExecIOSRB	srb;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = RF_DIR_NOXFER;
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 6;
	srb.cdb [0] = 0x00;
#ifdef SCSI_1
	srb.cdb [1] = lun << 5;
#endif
	return ASPIPerform (&srb);
}

int		ReadCapacity	(int AdapterNo, int scsiid, int lun, DWORD far *buf)

{
	struct ExecIOSRB	srb;
	int	result;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = RF_DIR_TOHOST;
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	srb.DataAllocLength = sizeof (DWORD)*2;
	srb.DataBufPtr = (BYTE far *) buf;
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 10;
	srb.cdb [0] = 0x25;
#ifdef SCSI_1
	srb.cdb [1] = lun << 5;
#endif
	if ((result = ASPIPerform (&srb)) == SUCCESS)
	{
		buf [0] = ntohl (buf [0]);
		buf [1] = ntohl (buf [1]);
	}
	return result;
}

int		StartStopUnit	(int AdapterNo, int scsiid, int lun, int flag)

{
	struct ExecIOSRB	srb;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = RF_DIR_NOXFER;
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 6;
	srb.cdb [0] = 0x1B;
#ifdef SCSI_1
	srb.cdb [1] = lun << 5;
#endif
	srb.cdb [4] = flag;
	return ASPIPerform (&srb);
}

int		PreventAllowMediumRemoval	(int AdapterNo, int scsiid, int lun, 
									int flag)

{
	struct ExecIOSRB	srb;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = RF_DIR_NOXFER;
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 6;
	srb.cdb [0] = 0x1E;
#ifdef SCSI_1
	srb.cdb [1] = lun << 5;
#endif
	srb.cdb [4] = flag;
	return ASPIPerform (&srb);
}

static int	near _ReadWrite10
	(int AdapterNo, int scsiid, int lun, DWORD lba, 
	WORD count, BYTE far *buf, int func)

{
	struct ExecIOSRB	srb;
	
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = (func ? RF_DIR_TOTARGET : RF_DIR_TOHOST);
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	srb.DataAllocLength = blocksize * (func == 3 ? 1 : count);
	srb.DataBufPtr = buf;
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 10;
	switch (func)
	{
		case 0:	srb.cdb [0] = 0x28;	break;
		case 1:	srb.cdb [0] = 0x2A;	break;
		case 2:	srb.cdb [0] = 0x2E;	break;
		case 3:	srb.cdb [0] = 0x41;	break;
	}
#ifdef SCSI_1
	srb.cdb [1] = lun << 5;
#endif
	srb.cdb [2] = (BYTE) (lba >> 24);
	srb.cdb [3] = (BYTE) (lba >> 16);
	srb.cdb [4] = (BYTE) (lba >>  8);
	srb.cdb [5] = (BYTE)  lba;
	srb.cdb [7] = count >> 16;
	srb.cdb [8] = count;
	return ASPIPerform (&srb);
}

int		Read10			(int AdapterNo, int scsiid, int lun, DWORD lba, 
						WORD count, BYTE far *buf)

{
	return _ReadWrite10 (AdapterNo, scsiid, lun, lba, count, buf, 0);
}


int		Write10			(int AdapterNo, int scsiid, int lun, DWORD lba, 
						WORD count, BYTE far *buf)

{
	return _ReadWrite10 (AdapterNo, scsiid, lun, lba, count, buf, 1);
}

int		WriteAndVerify	(int AdapterNo, int scsiid, int lun, DWORD lba, 
						WORD count, BYTE far *buf)

{
	return _ReadWrite10 (AdapterNo, scsiid, lun, lba, count, buf, 2);
}

int		WriteSame		(int AdapterNo, int scsiid, int lun, DWORD lba, 
						WORD count, BYTE far *buf)

{
	return _ReadWrite10 (AdapterNo, scsiid, lun, lba, count, buf, 3);
}

int		ModeSense	(int AdapterNo, int scsiid, int lun, 
					BYTE far *buf, int buflen)

{
	struct ExecIOSRB	srb;
	
	if  (buflen > 255) buflen = 255;
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = RF_DIR_TOHOST;
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	srb.DataAllocLength = buflen;
	srb.DataBufPtr = buf;
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 6;
	srb.cdb [0] = 0x1A;
#ifdef SCSI_1
	srb.cdb [1] = lun << 5;
#endif
	srb.cdb [2] = 0x3F;
	srb.cdb [4] = buflen;
	return ASPIPerform (&srb);
}

int		RequestSense	(int AdapterNo, int scsiid, int lun, 
						BYTE far *buf, int buflen)

{
	struct ExecIOSRB	srb;
	int		stat;
	
	if  (buflen > 255) buflen = 255;
	memset (&srb, 0, sizeof srb);
	srb.CommandCode = ASPI_EXEC_SCSI_IO;
	srb.AdapterNo = AdapterNo;
	srb.SCSIRequestFlags = RF_DIR_TOHOST;
	srb.TargetSCSIID = scsiid;
	srb.TargetLUN = lun;
	srb.DataAllocLength = buflen;
	srb.DataBufPtr = buf;
	srb.SenseAllocLength = SENSE_ALLOC_MAX;
	srb.CDBLength = 6;
	srb.cdb [0] = 0x03;
#ifdef SCSI_1
	srb.cdb [1] = lun << 5;
#endif
	srb.cdb [4] = buflen;
	if ((stat = ASPIPerform (&srb)) == SUCCESS)
	{
		SCSISenseKey   = buf [ 2] & 0x0F;
		SCSISenseCode  = buf [12];
		SCSISenseCodeQ = buf [13];
	}
	return stat;
}

