/*
	FreeBSD 2.x
	SCSI device driver interface module
	Copyright 1996 (C) ITO Takayuki
*/

#include	<stddef.h>
#include	<stdio.h>
#include	<sys/param.h>
#include	<sys/scsiio.h>
#include	"typedef.h"
#include	"fbsdscsi.h"

BYTE	SCSIDriverStatus = 0;
BYTE	SCSIStatus = S_GOOD;
BYTE	SCSISenseKey = 0;
BYTE	SCSISenseCode = 0;
BYTE	SCSISenseCodeQ = 0;
BOOL	UnitAtnRetry = TRUE;
DWORD	blocksize = 512U;

static int	SCSIPerform	(int fd, scsireq_t *psrq)
{
	int		stat;

	if (psrq->timeout == 0)
		psrq->timeout = 2*1000;
	psrq->senselen = SENSEBUFLEN;
	ioctl (fd, SCIOCCOMMAND, psrq);
	if ((SCSIDriverStatus = psrq->retsts) != SCCMD_OK)
		stat = FAILURE;
	else if ((SCSIStatus = psrq->status) == S_GOOD)
		stat = SUCCESS;
	else
	{
		stat = FAILURE;
		if (SCSIStatus == S_CHECK)
		{
			SCSISenseKey   = psrq->sense [ 2] & 0x0F;
			SCSISenseCode  = psrq->sense [12];
			SCSISenseCodeQ = psrq->sense [13];
		}
	}
	return stat;
}

int		_ReadWrite10	(int fd, DWORD lba, WORD count, void *buf, int func)

{
	scsireq_t	scsireq;

	bzero(&scsireq, sizeof (scsireq_t));
	switch (func)
	{
		case 0:	scsireq.cmd [0] = 0x28;	break;
		case 1:	scsireq.cmd [0] = 0x2A;	break;
		case 2:	scsireq.cmd [0] = 0x2E;	break;
		case 3:	scsireq.cmd [0] = 0x41;	break;
	}
	scsireq.cmd [2] = lba >> 24;
	scsireq.cmd [3] = lba >> 16;
	scsireq.cmd [4] = lba >>  8;
	scsireq.cmd [5] = lba;
	scsireq.cmd [7] = count >> 8;
	scsireq.cmd [8] = count;
	scsireq.cmdlen = 10;
	scsireq.databuf = buf;
	scsireq.datalen = blocksize * (func == 3 ? 1 : count);
	scsireq.flags = (func ? SCCMD_WRITE : SCCMD_READ);
	return SCSIPerform (fd, &scsireq);
}

int		TestUnitReady	(int fd)

{
	scsireq_t	scsireq;

	bzero(&scsireq, sizeof (scsireq_t));
	scsireq.cmd [0] = 0x00;
	scsireq.cmdlen = 6;
	scsireq.databuf = NULL;
	scsireq.datalen = 0;
	scsireq.flags = 0;
	return SCSIPerform (fd, &scsireq);
}

int		StartStopUnit	(int fd, int flag)

{
	scsireq_t	scsireq;

	bzero(&scsireq, sizeof (scsireq_t));
	scsireq.cmd [0] = 0x1B;
	scsireq.cmd [4] = flag;
	scsireq.cmdlen = 6;
	scsireq.databuf = NULL;
	scsireq.datalen = 0;
	scsireq.flags = 0;
	scsireq.timeout = 10*1000;
	return SCSIPerform (fd, &scsireq);
}

int		PreventAllowMediumRemoval	(int fd, int flag)

{
	scsireq_t	scsireq;

	bzero(&scsireq, sizeof (scsireq_t));
	scsireq.cmd [0] = 0x1E;
	scsireq.cmd [4] = flag;
	scsireq.cmdlen = 6;
	scsireq.databuf = NULL;
	scsireq.datalen = 0;
	scsireq.flags = 0;
	return SCSIPerform (fd, &scsireq);
}

int		ModeSense		(int fd, BYTE *buf, int buflen)

{
	scsireq_t	scsireq;

	if  (buflen > 255) buflen = 255;
	bzero(&scsireq, sizeof (scsireq_t));
	scsireq.cmd [0] = 0x1A;
	scsireq.cmd [2] = 0x3F;
	scsireq.cmd [4] = buflen;
	scsireq.cmdlen = 6;
	scsireq.databuf = buf;
	scsireq.datalen = buflen;
	scsireq.flags = SCCMD_READ;
	return SCSIPerform (fd, &scsireq);
}

int		Inquiry	(int fd, BYTE *buf, int buflen)

{
	scsireq_t	scsireq;

	if  (buflen > 255) buflen = 255;
	bzero(&scsireq, sizeof (scsireq_t));
	scsireq.cmd [0] = 0x12;
	scsireq.cmd [4] = buflen;
	scsireq.cmdlen = 6;
	scsireq.databuf = buf;
	scsireq.datalen = buflen;
	scsireq.flags = SCCMD_READ;
	return SCSIPerform (fd, &scsireq);
}

int		ReadCapacity	(int fd, DWORD *buf)

{
	int		result;
	scsireq_t	scsireq;

	bzero(&scsireq, sizeof (scsireq_t));
	scsireq.cmd [0] = 0x25;
	scsireq.cmdlen = 10;
	scsireq.databuf = (caddr_t) buf;
	scsireq.datalen = 8;
	scsireq.flags = SCCMD_READ;
	if ((result = SCSIPerform (fd, &scsireq)) == SUCCESS)
	{
		buf [0] = ntohl (buf [0]);
		buf [1] = ntohl (buf [1]);
	}
	return result;
}

int		FormatUnit		(int fd)

{
	scsireq_t	scsireq;

	bzero(&scsireq, sizeof (scsireq_t));
	scsireq.cmd [0] = 0x04;
	scsireq.cmdlen = 6;
	scsireq.databuf = NULL;
	scsireq.datalen = 0;
	scsireq.flags = 0;
	scsireq.timeout = 60*60*1000;
	return SCSIPerform (fd, &scsireq);
}

