/*
	HP-UX 9.0
	SCSI device driver interface module
	Copyright 1995,1996 (C) ITO Takayuki
*/

#include	<string.h>
#include	<sys/ioctl.h>
#include	<sys/scsi.h>
#include	<sys/types.h>
#include	<netinet/in.h>
#include	"typedef.h"
#include	"hpuxscsi.h"

unsigned	SCSIDriverStatus = 0;
BYTE	SCSIStatus = S_GOOD;
BYTE	SCSISenseKey = 0;
BYTE	SCSISenseCode = 0;
BYTE	SCSISenseCodeQ = 0;
BOOL	UnitAtnRetry = TRUE;
DWORD	blocksize = 512U;

static int	SCSIPerform	(int fd, struct sctl_io *psio)

{
	union sense_data	*psen;
	int		stat;
	
	ioctl (fd, SIOC_IO, psio);
	if ((SCSIDriverStatus = psio->cdb_status & 0xFFFFFF00) != 0)
		stat = FAILURE;
	else if ((SCSIStatus = psio->cdb_status & 0xFF) == S_GOOD)
		stat = SUCCESS;
	else
	{
		stat = FAILURE;
		if (SCSIStatus == S_CHECK_CONDITION)
		{
			psen = (union sense_data *) psio->sense;
			SCSISenseKey   = psen->r_sense2a.key;
			SCSISenseCode  = psen->r_sense2a.code;
			SCSISenseCodeQ = psen->r_sense2a.qualifier;
			if (SCSISenseKey == S_UNIT_ATTENTION && UnitAtnRetry)
				stat = SCSIPerform (fd, psio);
		}
	}
	return stat;
}

int		_ReadWrite10	(int fd, DWORD lba, WORD count, void *buf, int func)

{
	static struct sctl_io	sio;

	memset (&sio, 0, sizeof sio);
	sio.flags = (func ? 0 : SCTL_READ);
	sio.cdb_length = 10;
	switch (func)
	{
		case 0:	sio.cdb [0] = CMDread10;		break;
		case 1:	sio.cdb [0] = CMDwrite10;		break;
		case 2:	sio.cdb [0] = CMDwriteNverify;	break;
		case 3:	sio.cdb [0] = 0x41;				break;
	}
	sio.cdb [2] = lba >> 24;
	sio.cdb [3] = lba >> 16;
	sio.cdb [4] = lba >>  8;
	sio.cdb [5] = lba;
	sio.cdb [7] = count >> 8;
	sio.cdb [8] = count;
	sio.data = buf;
	sio.data_length = blocksize * (func == 3 ? 1 : count);
	sio.max_msecs = 1000;
	return SCSIPerform (fd, &sio);
}

int		TestUnitReady	(int fd)

{
	static struct sctl_io	sio;

	memset (&sio, 0, sizeof sio);
	sio.flags = 0;
	sio.cdb_length = 6;
	sio.cdb [0] = CMDtest_unit_ready;
	sio.data = NULL;
	sio.data_length = 0;
	sio.max_msecs = 100;
	return SCSIPerform (fd, &sio);
}

int		StartStopUnit	(int fd, int flag)

{
	static struct sctl_io	sio;

	memset (&sio, 0, sizeof sio);
	sio.flags = 0;
	sio.cdb_length = 6;
	sio.cdb [0] = CMDstart_stop_unit;
	sio.cdb [4] = flag;
	sio.data = NULL;
	sio.data_length = 0;
	sio.max_msecs = 100;
	return SCSIPerform (fd, &sio);
}

int		PreventAllowMediumRemoval	(int fd, int flag)

{
	static struct sctl_io	sio;

	memset (&sio, 0, sizeof sio);
	sio.flags = 0;
	sio.cdb_length = 6;
	sio.cdb [0] = CMDprevent_allow_media;
	sio.cdb [4] = flag;
	sio.data = NULL;
	sio.data_length = 0;
	sio.max_msecs = 100;
	return SCSIPerform (fd, &sio);
}

int		ModeSense		(int fd, BYTE *buf, int buflen)

{
	static struct sctl_io	sio;

	if  (buflen > 255) buflen = 255;
	memset (&sio, 0, sizeof sio);
	sio.flags = SCTL_READ;
	sio.cdb_length = 6;
	sio.cdb [0] = CMDmode_sense;
	sio.cdb [2] = 0x3F;
	sio.cdb [4] = buflen;
	sio.data = buf;
	sio.data_length = buflen;
	sio.max_msecs = 100;
	return SCSIPerform (fd, &sio);
}

int		ReadCapacity	(int fd, DWORD *buf)

{
	static struct sctl_io	sio;
	int	result;

	memset (&sio, 0, sizeof sio);
	sio.flags = SCTL_READ;
	sio.cdb_length = 10;
	sio.cdb [0] = CMDread_capacity;
	sio.data = buf;
	sio.data_length = 8;
	sio.max_msecs = 100;
	if ((result = SCSIPerform (fd, &sio)) == SUCCESS)
	{
		buf [0] = ntohl (buf [0]);
		buf [1] = ntohl (buf [1]);
	}
	return result;
}

int		FormatUnit		(int fd)

{
#if 1
	static struct sctl_io	sio;

	memset (&sio, 0, sizeof sio);
	sio.flags = 0;
	sio.cdb_length = 6;
	sio.cdb [0] = CMDformat_unit;
	sio.data = NULL;
	sio.data_length = 0;
	sio.max_msecs = 100;
	return SCSIPerform (fd, &sio);
#else
	struct sioc_format	fmt;
	
	fmt.fmt_optn = 0;
	fmt.interleave = 0;
	return ioctl (fd, SIOC_FORMAT, &fmt);
#endif
}

