/*
	Linux
	SCSI device driver interface module
	Copyright 1996 (C) ITO Takayuki

	Warning:
	This code is highly experimental.
	Error checking is neglected.
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<string.h>
#include	<fcntl.h>
#include	<errno.h>
#ifdef	LINUX_USE_IOCTL
	#include	<sys/ioctl.h>
#if 0
	#include	<linux/hdreg.h>
	#include	<linux/blkdev.h>
#endif
	#include	<linux/../scsi/scsi.h>
	#ifndef	SCSI_IOCTL_SEND_COMMAND
		#define	SCSI_IOCTL_SEND_COMMAND	1
	#endif
#else
	#include	<linux/../scsi/scsi.h>
	#include	<linux/../scsi/sg.h>
	#define	SCSI_OFF	sizeof(struct sg_header)
#endif
#include	"typedef.h"
#include	"linuscsi.h"
#define max(a,b) ((a)>(b) ? (a) : (b))

BYTE	SCSIDriverStatus = 0;
BYTE	SCSIStatus = GOOD;
BYTE	SCSISenseKey = 0;
BYTE	SCSISenseCode = 0;
BYTE	SCSISenseCodeQ = 0;
BOOL	UnitAtnRetry = TRUE;
DWORD	blocksize = 512U;

static int	SCSIPerform	(int fd,
	int cdb_len, BYTE *cdb,
	int in_size, BYTE *i_buff,
	int out_size, BYTE *o_buff)
#ifdef LINUX_USE_IOCTL
{
	int		stat, tmpbuflen;
	BYTE	*tmpbuf;

	tmpbuflen = sizeof (int) * 2 + max (cdb_len + in_size, out_size);
	tmpbuf = (BYTE *) malloc (tmpbuflen);

	*(int *)tmpbuf = in_size;
	*(int *)&tmpbuf [4] = out_size;

	memcpy (&tmpbuf [8], cdb, cdb_len);
	if (i_buff != NULL && in_size > 0)
		memcpy (&tmpbuf [8 + cdb_len], i_buff, in_size);

	stat = ioctl (fd, SCSI_IOCTL_SEND_COMMAND, tmpbuf);
	if (stat == FAILURE)
	{
		SCSIDriverStatus = errno;
		free (tmpbuf);
		return FAILURE;
	}

	if (o_buff != NULL && out_size > 0)
		memcpy (o_buff, &tmpbuf [8], out_size);

	free (tmpbuf);
	return stat;
}
#else
{
	int		stat, tmpbuflen;
	BYTE	*tmpbuf;
	struct sg_header	*sg_hd;

	tmpbuflen = SCSI_OFF + max (cdb_len + in_size, out_size);
	if (tmpbuflen > SG_BUFF_MAX)
	{
		SCSIDriverStatus = 0;
		return FAILURE;
	}

	tmpbuf = (BYTE *) malloc (tmpbuflen);
	sg_hd = (struct sg_header *) tmpbuf;

	memset (sg_hd, 0, SCSI_OFF);
#if 0
	sg_hd->pack_len    = SCSI_OFF + cdb_len + in_size;
#endif
	sg_hd->reply_len   = SCSI_OFF + out_size;
	sg_hd->twelve_byte = cmd_len == 12;

	memcpy (tmpbuf + SCSI_OFF, cdb, cdb_len);
	if (i_buff != NULL && in_size > 0)
		memcpy (tmpbuf + SCSI_OFF + cdb_len, i_buff, in_size);

	stat = write (fd, tmpbuf, SCSIOFF + cdb_len + in_size);
	if (stat != SCSIOFF + cdb_len + in_size || sg_hd->result != 0)
	{
		SCSIDriverStatus = errno;
		free (tmpbuf);
		return FAILURE;
	}

	stat = read (fd, tmpbuf, SCSI_OFF + out_size);
	if (stat != SCSIOFF + out_size || sg_hd->result != 0)
	{
		SCSIDriverStatus = sg_hd->result;	/* errnoƓ? */
		free (tmpbuf);
		return FAILURE;
	}

	if (o_buff != NULL && out_size > 0)
		memcpy (o_buff, tmpbuf + SCSI_OFF, out_size);

	SCSIDriverStatus = 0;
	if ((SCSIStatus = status_byte (hd_status)) == GOOD)
		stat = SUCCESS;
	else
	{
		stat = FAILURE;
		if (SCSIStatus == CHECK_CONDITION)
		{
			SCSISenseKey   = sg_hd->sense_buffer [ 2] & 0x0F;
			SCSISenseCode  = sg_hd->sense_buffer [12];
			SCSISenseCodeQ = sg_hd->sense_buffer [13];
		}
	}
	free (tmpbuf);
	return stat;
}
#endif

int		_ReadWrite10	(int fd, DWORD lba, WORD count, void *buf, int func)

{
	BYTE	cdb [10];

	memset (cdb, 0, sizeof cdb);
	switch (func)
	{
		case 0:	cdb [0] = 0x28;	break;	/* Read 10 */
		case 1:	cdb [0] = 0x2A;	break;	/* Write 10 */
		case 2:	cdb [0] = 0x2E;	break;	/* Write&Verify 10 */
		case 3:	cdb [0] = 0x41;	break;	/* Write Same 10 */
	}
	cdb [2] = lba >> 24;
	cdb [3] = lba >> 16;
	cdb [4] = lba >>  8;
	cdb [5] = lba;
	cdb [7] = count >> 8;
	cdb [8] = count;
	if (func)
		return SCSIPerform (fd, sizeof cdb, cdb,
			0, NULL, blocksize * (func == 3 ? 1 : count), buf);
	else
		return SCSIPerform (fd, sizeof cdb, cdb,
			blocksize * count, buf, 0, NULL);
}

int		TestUnitReady	(int fd)

{
	BYTE	cdb [6];

	memset (cdb, 0, sizeof cdb);
	cdb [0] = 0x00;
	return SCSIPerform (fd, sizeof cdb, cdb, 0, NULL, 0, NULL);
}

int		StartStopUnit	(int fd, int flag)

{
	BYTE	cdb [6];

	memset (cdb, 0, sizeof cdb);
	cdb [0] = 0x1B;
	cdb [4] = flag;
	return SCSIPerform (fd, sizeof cdb, cdb, 0, NULL, 0, NULL);
}

int		PreventAllowMediumRemoval	(int fd, int flag)

{
	BYTE	cdb [6];

	memset (cdb, 0, sizeof cdb);
	cdb [0] = 0x1E;
	cdb [4] = flag;
	return SCSIPerform (fd, sizeof cdb, cdb, 0, NULL, 0, NULL);
}

int		ModeSense		(int fd, BYTE *buf, int buflen)

{
	BYTE	cdb [6];

	if  (buflen > 255) buflen = 255;
	memset (cdb, 0, sizeof cdb);
	cdb [0] = 0x1A;
	cdb [2] = 0x3F;
	cdb [4] = buflen;
	return SCSIPerform (fd, sizeof cdb, cdb, 0, NULL, buflen, buf);
}

int		Inquiry	(int fd, BYTE *buf, int buflen)

{
	BYTE	cdb [6];

	if  (buflen > 255) buflen = 255;
	memset (cdb, 0, sizeof cdb);
	cdb [0] = 0x12;
	cdb [4] = buflen;
	return SCSIPerform (fd, sizeof cdb, cdb, 0, NULL, buflen, buf);
}

int		ReadCapacity	(int fd, DWORD *buf)

{
	int		result;
	BYTE	cdb [10];

	memset (cdb, 0, sizeof cdb);
	cdb [0] = 0x25;
	if ((result = SCSIPerform (fd, sizeof cdb, cdb, 0, NULL, 8, (BYTE *)buf))
		== SUCCESS)
	{
		buf [0] = ntohl (buf [0]);
		buf [1] = ntohl (buf [1]);
	}
	return result;
}

int		FormatUnit		(int fd)

{
	BYTE	cdb [6];

	memset (cdb, 0, sizeof cdb);
	cdb [0] = 0x04;
	return SCSIPerform (fd, sizeof cdb, cdb, 0, NULL, 0, NULL);
}

