/*
	mei - 3.5" MO disk formatter
	Copyright 1995,1996 (C) ITO Takayuki
	original for HP-UX
	ported for MS-DOS, FreeBSD, SunOS
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<fcntl.h>
#include	<errno.h>
#include	<ctype.h>
#include	<string.h>
#ifdef sunos4
	#undef	MB_LEN_MAX
#endif
#include	<limits.h>

#ifdef MSDOS
	#include	<dos.h>
	#include	<conio.h>
	#include	<io.h>
	#include	"typedef.h"
	#include	"aspiscsi.h"
	#include	"getopt.h"
	#include	"misclib.h"
	#ifdef NLS
		#include	"locale.h"
		#include	"msgcat.h"
		#define	CATFLAG	NL_CAT_LOCALE
		#define	NL_SETN	1
	#else
		#define	catgets(i,sn,mn,s)	(s)
	#endif
	#define	CLOSE(f)
#endif

#ifdef unix
	#include	<unistd.h>
	#if defined sunos4 || defined freebsd || defined linux
		#define	max(a,b)	((a)>(b)?(a):(b))
		#define	min(a,b)	((a)<(b)?(a):(b))
	#else
		#include	<macros.h>
	#endif
	#include	<time.h>
	#include	<sys/types.h>
	#include	<sys/param.h>
	#ifndef FILENAME_MAX
		#define	FILENAME_MAX	MAXPATHLEN
	#endif
	#include	<netinet/in.h>
	#ifdef hpux
		#include	<sys/ioctl.h>
		#include	<sys/scsi.h>
		#include	<sys/diskio.h>
	#elif defined osf1 || defined sunos4 || defined freebsd || defined linux
		#include	<sys/time.h>
	#endif
	#ifdef NLS
		#include	<locale.h>
		#include	<nl_types.h>
		#ifdef hpux
			#include	<msgcat.h>
			#define	CATFLAG	0
		#elif defined osf1
			#define	CATFLAG	NL_CAT_LOCALE
		#else
			#define	CATFLAG	0
		#endif
		#define	NL_SETN	1
	#else
		#define	catgets(i,sn,mn,s)	(s)
	#endif
	#include	"typedef.h"
	#ifdef hpux
		#include	"hpuxscsi.h"
		#define	CLOSE(f)	(UnlockDevice(f),close(f))
		#define	OPENFLAG	O_RDWR
	#elif defined osf1
		#include	"osf1scsi.h"
		#define	CLOSE(f)	CloseCAM()
	#elif defined sunos4
		#include	"sunscsi.h"
		#include	"getopt.h"
		#include	"errno.h"
		#define	CLOSE(f)	close(f)
		#define	OPENFLAG	(O_RDONLY|O_NDELAY)
	#elif defined sunos5
		#include	"sunscsi.h"
		#include	"errno.h"
		#define	CLOSE(f)	close(f)
		#define	OPENFLAG	(O_RDONLY|O_NDELAY)	/* O_RDWR rejects R/O medium */
	#elif defined freebsd
		#include	"fbsdscsi.h"
		#define	CLOSE(f)	close(f)
		#define	OPENFLAG	O_RDWR
	#elif defined linux
		#include	"linuscsi.h"
		#define	CLOSE(f)	close(f)
		#define	OPENFLAG	O_RDWR
	#endif
	#define	O_BINARY	0x0000		/* dummy */
#endif

#include	"m_mei.h"

#define	NAME	"mei"

#ifdef unix
#undef	DOS_BULK_FAT_WRITE
#endif

#if defined MSDOS || defined osf1
#define	USE_SCSI_WRITE
#endif

/* writing multibyte data at any address in little endian */
#if defined MSDOS || defined freebsd
	#define	POKEB(a,n)	(*( BYTE *)(a)=(n))
	#define	POKEW(a,n)	(*( WORD *)(a)=(n))
	#define	POKED(a,n)	(*(DWORD *)(a)=(n))
#else	/* aligned CPUs */
	#define	POKEB(a,n)	(((BYTE *)(a))[0] =  (n)        & 255)
	#define	POKEW(a,n)	(((BYTE *)(a))[0] =  (n)        & 255, \
						 ((BYTE *)(a))[1] = ((n) >>  8) & 255)
	#define	POKED(a,n)	(((BYTE *)(a))[0] =  (n)        & 255, \
						 ((BYTE *)(a))[1] = ((n) >>  8) & 255, \
						 ((BYTE *)(a))[2] = ((n) >> 16) & 255, \
						 ((BYTE *)(a))[3] = ((n) >> 24) & 255)
#endif

#define	MEDIUM_TYPES				5	/* 128MB, 230MB, 385MB, 540MB, 640MB */
#define	MAX_MO_DENSITY				5
#define	SEMIIBM_MEDIUM_TYPES		2	/* 128MB, 230MB */
#define	SEMIIBM_MAX_MO_DENSITY		2

#define	TOTAL_SECTORS_MO1D	((10000UL - 3*2) * 25 - 1024)
#define	TOTAL_SECTORS_MO2D	((17940UL - 3*2 - 2*2*10 - 41) * 25)
#define	TOTAL_SECTORS_MO3D	0				/* unknown */
#define	TOTAL_SECTORS_MO4D	(41660UL * 25)	/* preliminary */
#define	TOTAL_SECTORS_MO5D	310352UL

#define	TOTAL_SECTORS_PD	((20400UL - 3*2 - 2*2*10 - 65) * 64)
#define	TOTAL_SECTORS_HS1	318988UL

#define	MAX_TOTAL_SECTORS_MO1D	(TOTAL_SECTORS_MO1D + 1024)

static const struct PhysicalParams_t
{
	DWORD	MinSectors, MaxSectors;
	WORD	SectorSize, SectorsPerLogicalTrack, SectorStep;
} PhysicalParams [MEDIUM_TYPES] =
{
	/* 128MB */
	{ TOTAL_SECTORS_MO1D, MAX_TOTAL_SECTORS_MO1D, 
	  512, 25, 25*5 },
	/* 230MB */
	{ TOTAL_SECTORS_MO2D, TOTAL_SECTORS_MO2D,
	  512, 25, 30*4 },
	/* 385MB (unknown) */
	{ TOTAL_SECTORS_MO3D, TOTAL_SECTORS_MO3D,
	  512, 25, 25*5 },
	/* 540MB */
	{ TOTAL_SECTORS_MO4D, TOTAL_SECTORS_MO4D,
	  512, 25, 25*5 },
	/* 640MB */
	{ TOTAL_SECTORS_MO5D, TOTAL_SECTORS_MO5D,
	  2048, 32, 25*1 },

#if 0
	/* PD */
	{ TOTAL_SECTORS_PD, TOTAL_SECTORS_PD,
	  512, 32, 66*1 }
	/* HS-1 */
	{ TOTAL_SECTORS_HS1, TOTAL_SECTORS_HS1,
	  2048, 32, 31 }
#endif
};

#define	DIRENT_SIZE	32U
#define	FATS		2
#define	HIDDEN_SECTORS	0

#define	MDB			0xF0
#define	FATID		0xF0
#define	FA_LABEL	0x08
#define	IPL_FILLER	0x00
#define	FAT_FILLER	0x00
#define	DIR_FILLER	0xE5

#define	ASC_WRITE_PROTECT	0x27

#define	BELL		"\a\a\a\a\a"
#define	BOOTFILEEXT	".ipl"

enum boottype_t { NOBOOT, OWNBOOT, EXTBOOT };
enum fmttype_t { IBMSTD, SEMIIBM };

static const struct LogicalParams_t
{
	WORD	SectorsPerCluster, ReservedSectors, SectorsPerFAT;
	DWORD	LogicalSectors;
}
LogicalParams [MEDIUM_TYPES + SEMIIBM_MEDIUM_TYPES] =
{
	{  4, 1, 243, TOTAL_SECTORS_MO1D - 26 },	/* 128MB, IBMstd */
	{  8, 1, 218, TOTAL_SECTORS_MO2D },			/* 230MB, IBMstd */
	{  0, 1,   0, 0 },							/* 385MB, IBMstd (unknown) */
	{ 16, 1, 255, TOTAL_SECTORS_MO4D },			/* 540MB, IBMstd */
	{  8, 1,  38, TOTAL_SECTORS_MO5D - 3152 },	/* 640MB, IBMstd */

	{  8, 4, 124, TOTAL_SECTORS_MO1D - 2 },		/* 128MB, semiIBM */
	{ 16, 8, 112, TOTAL_SECTORS_MO2D - 5 }		/* 230MB, semiIBM */

#if 0
	{ 32, 1, 159, TOTAL_SECTORS_PD },			/* PD */
	{  8, 1,  39, TOTAL_SECTORS_HS1 },			/* HS-1 (experimental) */
#endif
};

struct bootsector
{
	/* preamble */
	BYTE	JumpInst					[ 3];
	BYTE	OEMBanner					[ 8];
	
	/* BPB */
	BYTE	LogicalSectorSize			[ 2];
	BYTE	SectorsPerCluster			[ 1];
	BYTE	ReservedSectors				[ 2];
	BYTE	FATs						[ 1];
	BYTE	RootDirEntries				[ 2];
	BYTE	LogicalSectors16			[ 2];
	BYTE	MediaDescriptor				[ 1];
	BYTE	SectorsPerFAT				[ 2];
	
	BYTE	SectorsPerTrack				[ 2];
	BYTE	Heads						[ 2];
	BYTE	HiddenSectors				[ 4];
	
	/* extended boot record info. (DOS 4.0 and later) */
	BYTE	LogicalSectors32			[ 4];
	BYTE	PhysicalDriveNo				[ 1];
	BYTE	HeadNo						[ 1];
	BYTE	ExtendedSignature			[ 1];
	BYTE	VolumeSerialNo				[ 4];
	BYTE	VolumeLabel					[11];
	BYTE	FATType						[ 8];
	
	/* PC-9801 specific */
	BYTE	TopPhysicalSectorNo 		[ 4];
	BYTE	IOSYSPhysicalSectorOffset	[ 2];
	BYTE	PhysicalSectorSize			[ 2];
	BYTE	Reserved2					[ 1];
	
	/* IPL code follows */
	BYTE	IPLCode						[ 3];
};

int		SectorsPerLogicalTrack, RootDirEntries, RootDirSectors, 
		SectorsPerFAT, SectorsPerCluster, ReservedSectors, 
		SectorStep = 0, SectorSize, Heads = 64;
int		MediumType = 0,
		year = 0, month = 0, day = 0, 
		hour = 0, minute = 0, second = 0, msec = 0;
DWORD	LogicalSectors;
BOOL	errorf = FALSE, verbosef = TRUE, phyfrmf = FALSE, 
		fatf = TRUE, rootf = TRUE, labelf = FALSE, ejectf = TRUE,
		progressf = FALSE, debugf = TRUE;
WORD	msdosdate, msdostime;
char	bootfilepath [FILENAME_MAX], vlabel [40] = "NO NAME    ";

#ifdef MSDOS
	static int	ha_id = 0, scsi_id = -1, lun = 0;
	#define	FD	ha_id, scsi_id, lun
	static BYTE	far *iobuf;
	static WORD	iobufseg;
#else
	static BYTE	*iobuf;
	#ifdef osf1
		static int	ha_id = 0, scsi_id = -1, lun = 0;
		#define	FD	ha_id, scsi_id, lun
	#else
		static char	devicefile [] = DEVICE_FILE;
		static int	fd;
		#define	FD	fd
	#endif
#endif

#ifdef NLS
nl_catd	catd;
#endif

enum boottype_t	boottype = OWNBOOT;
enum fmttype_t	fmttype = IBMSTD;

void	banner	(void)

{
#ifdef MSDOS
	putchar ('\n');
#endif
	fprintf (stdout, catgets (catd, NL_SETN, 1, MSG_1), NAME, RELEASE);
}

void	usage	(void)

{
	fprintf (stderr, catgets (catd, NL_SETN, 4, MSG_4), NAME);
#ifndef SECURE
	fputs (catgets (catd, NL_SETN, 5, MSG_5), stderr);
#if defined MSDOS || defined osf1
	fputs (catgets (catd, NL_SETN, 6, MSG_6), stderr);
#else
	fputs (catgets (catd, NL_SETN, 7, MSG_7), stderr);
#endif
#endif
	putchar ('\n');
	fputs (catgets (catd, NL_SETN, 8, MSG_8), stderr);
#if defined MSDOS || defined osf1
	fputs (catgets (catd, NL_SETN, 17, MSG_17), stderr);
#else
	fprintf (stderr, catgets (catd, NL_SETN, 18, MSG_18), DEVICE_FILE);
#endif
}

void	StatusReport	(void)

{
	if (debugf)
	{
#ifdef MSDOS
		fprintf (stderr, "%02X %02X ",
			ASPICommandStatus, ASPIAdapterStatus);
#else
		fprintf (stderr, "%04X ", SCSIDriverStatus);
#endif
		fprintf (stderr, "%02X %02X %02X %02X\n",
			SCSIStatus, SCSISenseKey, SCSISenseCode, SCSISenseCodeQ);
	}
}

void	analyze_option	(int argc, char **argv)

{
	int		c, fd2;
	char	*p;
	
	while ((c = getopt (argc, argv, "12345isuUcvhqn:b:d:l:D")) != EOF &&
		c != '?' && !errorf)
	{
		switch (c)
		{
		case 'i':	fmttype  = IBMSTD;	break;
		case 's':	fmttype  = SEMIIBM;	break;
		case 'c':	phyfrmf  = FALSE;	break;
		case 'v':	verbosef = TRUE;	break;
		case 'q':	verbosef = FALSE;	break;
		case 'D':	debugf   = TRUE;	break;
			
		case 'u':
			phyfrmf = TRUE;
			progressf = FALSE;
			break;
			
		case 'U':
			phyfrmf = progressf = TRUE;
			break;
			
		case 'h':
			usage ();
			exit (EXIT_FAILURE);
			
		case 'n':
			for (p = optarg; (c = *p) != '\0'; p++)
			{
				switch (c)
				{
				case 'b':	boottype = NOBOOT;	break;
				case 'f':	fatf   = FALSE;		break;
				case 'd':
				case 'r':	rootf  = FALSE;		break;
				case 'e':	ejectf = FALSE;		break;
					
				default:
					errno = EINVAL;
					fprintf (stderr, "%s: %s -- %c\n", 
						NAME, strerror (errno), c);
					errorf = TRUE;
					break;
				}
			}
			break;
			
#ifndef SECURE
		case 'b':
			strcpy (bootfilepath, optarg);
			if ((fd2 = open (bootfilepath, O_RDONLY)) == FAILURE)
			{
				errno = ENOENT;
				fprintf (stderr, "%s: %s -- %s\n", 
					NAME, strerror (errno), optarg);
				errorf = TRUE;
			}
			else
			{
				boottype = EXTBOOT;
				close (fd2);
			}
			break;
			
		case 'd':
#if defined MSDOS || defined osf1
			if (strchr (optarg, ',') == NULL)
				scsi_id = atoi (optarg);
			else
				sscanf (optarg, "%d,%d,%d", &ha_id, &scsi_id, &lun);
#else
			if (optarg [0] == _PATH_DELIM)
				strcpy (devicefile, optarg);
			else
				sprintf (devicefile, DEVICE_FILE_TMPL, optarg);
#endif
			break;
#endif
			
		case 'l':
			strcpy (vlabel, optarg);
			strcat (vlabel, "          ");
			labelf = TRUE;
			break;

		default:
			if (isdigit (c))
			{
				if ((c -= '0') <= MAX_MO_DENSITY)
				{
					MediumType = c;
				}
			}
			break;
		}
	}
	
	if (fmttype == SEMIIBM && MediumType > SEMIIBM_MAX_MO_DENSITY)
	{
		errno = EINVAL;
		fprintf (stderr, "%s: %s\n", 
			NAME, strerror (errno));
		errorf = TRUE;
	}
	if (errorf || c == '?' || argc - optind != 0)
	{
		usage ();
		exit (EXIT_FAILURE);
	}
}

#ifdef hpux
int		LockDevice	(int fd)

{
	int		flag = 3;
	
	return ioctl (fd, SIOC_EXCLUSIVE, &flag);
}

int		UnlockDevice	(int fd)

{
	int		flag = 2;
	
	return ioctl (fd, SIOC_EXCLUSIVE, &flag);
}
#endif

void	open_device	(void)

{
	BYTE	InquiryData [255];
	#define	DEVTYPE	(InquiryData [0] & 0x1F)
	#define	RMB		(InquiryData [1] & 0x80)
	
#ifdef MSDOS
	if (InitASPI () == FAILURE)
	{
		fputs (catgets (catd, NL_SETN, 20, MSG_20), stderr);
		exit (EXIT_FAILURE);
	}
#else /* MSDOS */
#ifdef osf1
	if (OpenCAM () == FAILURE)
#else
	if ((fd = open (devicefile, OPENFLAG)) == FAILURE)
#endif
	{
		perror (NAME ": open()");
		exit (EXIT_FAILURE);
	}
#endif /* MSDOS */
	
	UnitAtnRetry = TRUE;
#if defined MSDOS || defined osf1
	if (scsi_id < 0)
	{
		for (scsi_id = 0; scsi_id <= 7; scsi_id++)
			if (Inquiry (FD, InquiryData, sizeof InquiryData) == SUCCESS && 
#ifdef MSDOS
				DEVTYPE == SCSI_MO)
#else
				(DEVTYPE == SCSI_MO || (DEVTYPE == SCSI_DIRECT_ACCESS && RMB)))
#endif
				break;
		if (scsi_id > 7)
		{
			fputs (catgets (catd, NL_SETN, 21, MSG_21), stderr);
			CLOSE (fd);
			exit (EXIT_FAILURE);
		}
		if (verbosef)
			printf (catgets (catd, NL_SETN, 22, MSG_22), scsi_id);
		return;
	}
#endif
	
#ifdef hpux
	LockDevice (fd);
#endif
	
	if (Inquiry (FD, InquiryData, sizeof InquiryData) == FAILURE)
	{
		fputs (catgets (catd, NL_SETN, 23, MSG_23), stderr);
		StatusReport ();
		CLOSE (fd);
		exit (EXIT_FAILURE);
	}
	
	if (DEVTYPE != SCSI_MO)
	{
		if (DEVTYPE == SCSI_DIRECT_ACCESS && RMB)
		{
#if defined MSDOS || defined hpux || defined freebsd
			char	lbuf [80], yes, no, c;
			
			yes = tolower (*catgets (catd, NL_SETN, 27, MSG_27));
			no  = tolower (*catgets (catd, NL_SETN, 28, MSG_28));
			fputs (catgets (catd, NL_SETN, 24, MSG_24), stdout);
			do
			{
				fputs (catgets (catd, NL_SETN, 25, MSG_25), stdout);
				fgets (lbuf, sizeof lbuf, stdin);
				c = tolower (*lbuf);
			} while (c != yes && c != no);
			if (c == no)
			{
				CLOSE (fd);
				exit (EXIT_FAILURE);
			}
#endif
		}
		else
		{
			fputs (catgets (catd, NL_SETN, 26, MSG_26), stderr);
			CLOSE (fd);
			exit (EXIT_FAILURE);
		}
	}
}

void	check_medium	(void)

{
	BYTE	mode_data [255];
	
#if defined MSDOS || defined freebsd
	TestUnitReady (FD);
#endif
	if (TestUnitReady (FD) == FAILURE)
	{
		if (verbosef)
		{
			fputs (catgets (catd, NL_SETN, 30, MSG_30), stdout);
			fputs (catgets (catd, NL_SETN, 31, MSG_31), stdout);
			putchar ('\n');
		}
		while (TestUnitReady (FD) == FAILURE)
#ifdef MSDOS
			kbhit ();
#endif
#ifdef unix
			sleep (0);
#endif
	}
	
	if (ModeSense (FD, mode_data, sizeof mode_data) == SUCCESS)
	{
		if (mode_data [2] & 0x80)		/* if write protected */
		{
			fputs (BELL, stderr);
			fputs (catgets (catd, NL_SETN, 34, MSG_34), stderr);
			if (verbosef)
				fputs (catgets (catd, NL_SETN, 35, MSG_35), stdout);
			StartStopUnit (FD, 2);
			CLOSE (fd);
			exit (EXIT_FAILURE);
		}
	}
	else
		StatusReport ();
}

void	setup_params	(void)

{
	DWORD	TotalSectors;
	DWORD	ReadCapacity_data [2];
	const struct PhysicalParams_t	*phyprm;
	const struct LogicalParams_t	*logprm;
	int	i;
	
	if (ReadCapacity (FD, ReadCapacity_data) == FAILURE)
	{
		fputs (catgets (catd, NL_SETN, 46, MSG_46), stderr);
		StatusReport ();
		CLOSE (fd);
		exit (EXIT_FAILURE);
	}
	
	TotalSectors =        ReadCapacity_data [0] + 1;
	SectorSize   = (WORD) ReadCapacity_data [1];
	if (MediumType == 0)
	{
		for (MediumType = 1; MediumType <= MAX_MO_DENSITY; MediumType++)
		{
			phyprm = &PhysicalParams [MediumType - 1];
			if (phyprm->MinSectors <= TotalSectors && 
				TotalSectors <= phyprm->MaxSectors)
				break;
		}
		if (MediumType > MAX_MO_DENSITY)
		{
			fprintf (stderr, catgets (catd, NL_SETN, 33, MSG_33), 
				(unsigned long int) TotalSectors);
			CLOSE (fd);
			exit (EXIT_FAILURE);
		}
	}
	else
		phyprm = &PhysicalParams [MediumType - 1];
	if (SectorSize != phyprm->SectorSize)
	{
		fprintf (stderr, catgets (catd, NL_SETN, 32, MSG_32), 
			(unsigned long int) SectorSize);
		CLOSE (fd);
		exit (EXIT_FAILURE);
	}
	blocksize = SectorSize;
#if defined MSDOS || defined freebsd
	SectorStep = phyprm->SectorStep;
#elif defined SUN_USE_GD
	SectorStep = 4;		/* gd driver rejects I/O of over 4 sectors? */
#elif defined linux
	SectorStep = SG_BUFF_MAX / SectorSize - 1/* space for header and CDB */;
#endif
	SectorsPerLogicalTrack = phyprm->SectorsPerLogicalTrack;
	
	i = MediumType - 1;
	if (fmttype == SEMIIBM)
		i += MEDIUM_TYPES;
	logprm = &LogicalParams [i];
	SectorsPerCluster = logprm->SectorsPerCluster;
	ReservedSectors   = logprm->ReservedSectors;
	SectorsPerFAT     = logprm->SectorsPerFAT;
	LogicalSectors    = logprm->LogicalSectors;
	
	RootDirEntries = 512;
	RootDirSectors = RootDirEntries * DIRENT_SIZE / SectorSize;
}

void	alloc_iobuf	(void)

{
#ifdef unix
	if ((iobuf = malloc (SectorSize * SectorsPerFAT)) == NULL)
#else
#ifdef DOS_USE_ALLOCMEM
	#ifdef DOS_BULK_FAT_WRITE
	if (allocmem (SectorSize / 16 * SectorsPerFAT, (unsigned *) &iobufseg) 
		!= ALLOCMEM_SUCCESS)
	#else
	if (allocmem (SectorSize / 16 * SectorStep, (unsigned *) &iobufseg) 
		!= ALLOCMEM_SUCCESS)
	#endif /* DOS_BULK_FAT_WRITE */
#else /* DOS_USE_ALLOCMEM */
	#ifdef DOS_BULK_FAT_WRITE
	if ((iobuf = (BYTE far *) malloc (SectorSize * SectorsPerFAT + 16)) 
		== NULL)
	#else
	if ((iobuf = (BYTE far *) malloc (SectorSize * SectorStep + 16)) == NULL)
	#endif /* DOS_BULK_FAT_WRITE */
#endif /* DOS_USE_ALLOCMEM */
#endif /* unix */
	{
		fputs (catgets (catd, NL_SETN, 36, MSG_36), stderr);
		CLOSE (fd);
		exit (EXIT_FAILURE);
	}
#if defined MSDOS && defined (DOS_USE_ALLOCMEM)
	iobuf = MK_FP (iobufseg, 0);
#endif
}

void	prompt1	(void)

{
	char	lbuf [80];
	
	if (verbosef)
	{
		fputs (catgets (catd, NL_SETN, 41, MSG_41), stdout);
		fputs (catgets (catd, NL_SETN, 43, MSG_43), stdout);
		fputs (catgets (catd, NL_SETN, 31, MSG_31), stdout);
		fgets (lbuf, sizeof lbuf, stdin);
	}
}

void	prompt2	(void)

{
	char	lbuf [80], lbuf2 [80];
	static char	*density_msg [MEDIUM_TYPES] = 
		{ MSG_60, MSG_61, MSG_62, MSG_63, MSG_64 };
	
	if (verbosef)
	{
		strcpy (lbuf, catgets (catd, NL_SETN, 60 + MediumType - 1, 
			density_msg [MediumType - 1]));
		switch (fmttype)
		{
		case IBMSTD:
			strcpy (lbuf2, catgets (catd, NL_SETN, 39, MSG_39));
			break;
			
		case SEMIIBM:
			strcpy (lbuf2, catgets (catd, NL_SETN, 40, MSG_40));
			break;
		}
		printf (catgets (catd, NL_SETN, 42, MSG_42), lbuf, lbuf2);
		fputs (catgets (catd, NL_SETN, 43, MSG_43), stdout);
		fputs (catgets (catd, NL_SETN, 31, MSG_31), stdout);
		fgets (lbuf, sizeof lbuf, stdin);
	}
}

#ifdef MSDOS
static int	format_progress		(void)

{
	int		prevrate, rate, stat;
	BYTE	sense_buf [18];
	static char		bar [] = "********************************";
	
	fputs ("+-------+-------+-------+-------+\n*", stdout);
	prevrate = 0;
	for (;;)
	{
		RequestSense (FD, sense_buf, 18);
		if (SCSIStatus != S_GOOD)	/* RequestSense itself failed */
		{
			stat = FAILURE;
			break;
		}
		if (SCSISenseKey   == 2 &&		/* Not Ready */
			SCSISenseCode  == 4 && 
			SCSISenseCodeQ == 4)		/* Format in progress */
		{
			if (sense_buf [15] & 0x80)	/* SKSV == 1 */
			{
				rate = sense_buf [16] >> 3;		/* progress indicator */
				if (rate > prevrate)
				{
					printf ("%.*s", rate - prevrate, bar);
					prevrate = rate;
				}
			}
#ifndef NO_SLEEP
			sleep (1);
#endif
		}
		else
		{
			stat = (SCSISenseKey == 0 ? SUCCESS : FAILURE);
			break;
		}
	}
	
	if (stat == SUCCESS)
		printf ("%.*s", 32 - prevrate, bar);
	putchar ('\n');
	
	return stat;
}
#endif

void	physical_format	(void)

{
	int		stat;
#ifdef MSDOS
	BYTE	fmt_param [4];
#endif
	
	if (verbosef)
		fputs (catgets (catd, NL_SETN, 44, MSG_44), stdout);
	PreventAllowMediumRemoval (FD, 1);
#ifdef MSDOS
	if (verbosef && progressf)
	{
		memset (fmt_param, 0, sizeof fmt_param);
		fmt_param [1] = 0x82;
		if (FormatUnit2 (FD, 0x10, 0, fmt_param, 4) 
			== SUCCESS)
			stat = format_progress ();
		else if (SCSISenseKey == 0x05)	/* Illegal Request */
			stat = FormatUnit (FD);
		else
			stat = FAILURE;
	}
	else
#endif
		stat = FormatUnit (FD);
	PreventAllowMediumRemoval (FD, 0);
	
	if (stat == FAILURE)
	{
		fputs (catgets (catd, NL_SETN, 45, MSG_45), stderr);
		StatusReport ();
		CLOSE (fd);
		exit (EXIT_FAILURE);
	}
}

void	setup_time	(void)

{
#ifdef MSDOS
	#ifdef __TURBOC__
	struct time	ltime;
	struct date	ldate;
	#else
	struct dostime_t	ltime;
	struct dosdate_t	ldate;
	#endif
#else
	struct timeval	unixtime;
	struct tm	*ltime;
#endif
	
#ifdef MSDOS
	#ifdef __TURBOC__
	gettime (&ltime);
	getdate (&ldate);
	year   = ldate.da_year;
	month  = ldate.da_mon;
	day    = ldate.da_day;
	hour   = ltime.ti_hour;
	minute = ltime.ti_min;
	second = ltime.ti_sec;
	msec   = ltime.ti_hund * 10;
	#else
	_dos_gettime (&ltime);
	_dos_getdate (&ldate);
	year   = ldate.year;
	month  = ldate.month;
	day    = ldate.day;
	hour   = ltime.hour;
	minute = ltime.minute;
	second = ltime.second;
	msec   = ltime.hsecond * 10;
	#endif
#else /* MSDOS */
	if (gettimeofday (&unixtime, NULL) == 0)
	{
		ltime = localtime ((time_t *)&unixtime.tv_sec);
		year   = ltime->tm_year + 1900;
		month  = ltime->tm_mon + 1;
		day    = ltime->tm_mday;
		hour   = ltime->tm_hour;
		minute = ltime->tm_min;
		second = ltime->tm_sec;
		msec   = unixtime.tv_usec / 1000;
	}
#endif /* MSDOS */
	msdosdate = ((year-1980) << 9) | (month << 5) | day;
	msdostime = (hour << 11) | (minute << 5) | (second >> 1);
}

void	build_mdt	(struct bootsector *bootsect)

{
	struct bootsector	tmpboot;
	int		disp, len;
	
	memcpy (&tmpboot.OEMBanner [0], NAME "     ", 5);
	memcpy (&tmpboot.OEMBanner [5], RELEASE, 3);
	POKEW (tmpboot.LogicalSectorSize, SectorSize);
	POKEB (tmpboot.SectorsPerCluster, SectorsPerCluster);
	POKEW (tmpboot.ReservedSectors, ReservedSectors);
	POKEB (tmpboot.FATs, FATS);
	POKEW (tmpboot.RootDirEntries, RootDirEntries);
	if ( LogicalSectors > 0xFFFFUL )
	{
		POKEW (tmpboot.LogicalSectors16, 0);
		POKED (tmpboot.LogicalSectors32, LogicalSectors);
	}
	else
	{
		POKEW (tmpboot.LogicalSectors16, (WORD) LogicalSectors);
		POKED (tmpboot.LogicalSectors32, 0L);
	}
	POKEB (tmpboot.MediaDescriptor, MDB);
	POKEW (tmpboot.SectorsPerFAT, SectorsPerFAT);
	POKEW (tmpboot.SectorsPerTrack, SectorsPerLogicalTrack);
	POKEW (tmpboot.Heads, Heads);
	POKED (tmpboot.HiddenSectors, HIDDEN_SECTORS);
	
	POKEB (tmpboot.PhysicalDriveNo, 0x80);
	POKEB (tmpboot.HeadNo, 0);
	POKEB (tmpboot.ExtendedSignature, 0x29);
	memcpy (tmpboot.FATType, "FAT16   ", 8);
	
	POKED (tmpboot.TopPhysicalSectorNo, 0L);
	POKEW (tmpboot.IOSYSPhysicalSectorOffset, 
		ReservedSectors + SectorsPerFAT * FATS + RootDirSectors);
	POKEW (tmpboot.PhysicalSectorSize, SectorSize);
	POKEB (tmpboot.Reserved2, 0);
	
	len = INT_MAX;
	if (bootsect->JumpInst [0] == 0xEB)			/* JMP SHORT */
	{
		disp = (INT8) bootsect->JumpInst [1];
		if (disp >= 1)		/* if forward jump */
			len = disp - 1;
	}
	else if (bootsect->JumpInst [0] == 0xE9)	/* JMP NEAR */
	{
		disp = (INT16) (bootsect->JumpInst [1] + 
			(bootsect->JumpInst [2] << 8));
		if (disp >= 0)
			len = disp;
	}
	len = min (len, (int)(tmpboot.IPLCode - tmpboot.OEMBanner));
	memcpy (bootsect->OEMBanner, tmpboot.OEMBanner, len);
}

void	build_dummyipl	(struct bootsector *bootsect)

{
	bootsect->JumpInst [0] = 0xEB;
	bootsect->JumpInst [1] = 
		(BYTE) (bootsect->IPLCode - &bootsect->JumpInst [2]);
	bootsect->JumpInst [2] = 0x90;
	bootsect->IPLCode [0] = 0xCB;		/* RETF */
}

void	build_boot	(void)

{
	int		fd2;
	char	*bootfilename;
	struct bootsector	*bootsect;
	
	memset ((char *) iobuf, IPL_FILLER, SectorSize * ReservedSectors);
	bootsect = (struct bootsector *) iobuf;
	switch (boottype)
	{
	case NOBOOT:
#ifdef USE_SCSI_WRITE
		if (Read10 (FD, 0UL, ReservedSectors, iobuf) == FAILURE)
			StatusReport ();
#else
		lseek (fd, 0L, SEEK_SET);
		if (read (fd, iobuf, SectorSize * ReservedSectors) <
			SectorSize * ReservedSectors)
			perror (NAME ": build_boot()");
#endif
		break;
		
	case EXTBOOT:
		if ((fd2 = open (bootfilepath, O_RDONLY | O_BINARY)) == FAILURE)
		{
			perror (NAME ": build_boot()");
			/* the only perror() call in MS-DOS version */
			CLOSE (fd);
			exit (EXIT_FAILURE);
		}
		if (verbosef)
			printf (catgets (catd, NL_SETN, 48, MSG_48), bootfilepath);
		read (fd2, iobuf, SectorSize * ReservedSectors);
		close (fd2);
		break;
		
	case OWNBOOT:
#ifdef MSDOS
		bootfilename = strrchr (bootfilepath, _PATH_DELIM) + 1;
#else
		bootfilename = bootfilepath + strlen (bootfilepath);
		if (*(bootfilename - 1) != _PATH_DELIM)
			*bootfilename++ = _PATH_DELIM;
#endif
		sprintf (bootfilename, NAME "%c%c" BOOTFILEEXT, 
			'0' + MediumType, (fmttype == IBMSTD ? 'i' : 's'));
		if ((fd2 = open (bootfilepath, O_RDONLY | O_BINARY)) != FAILURE)
		{
			if (verbosef)
				printf (catgets (catd, NL_SETN, 48, MSG_48), bootfilename);
			read (fd2, iobuf, SectorSize * ReservedSectors);
			close (fd2);
		}
		else
		{
			strcpy (bootfilename, NAME BOOTFILEEXT);
			if ((fd2 = open (bootfilepath, O_RDONLY | O_BINARY)) != FAILURE)
			{
				if (verbosef)
				{
					printf (catgets (catd, NL_SETN, 48, MSG_48), 
						NAME BOOTFILEEXT);
				}
				read (fd2, iobuf, SectorSize * ReservedSectors);
				close (fd2);
			}
			else
			{
				if (verbosef)
					fputs (catgets (catd, NL_SETN, 51, MSG_51), stdout);
				build_dummyipl (bootsect);
			}
		}
		break;
	}
	
	build_mdt (bootsect);
	if (bootsect->ExtendedSignature [0] == 0x29)
	{
		memcpy (bootsect->VolumeLabel, vlabel, 11);
		if (year == 0)
			POKED (bootsect->VolumeSerialNo, 0L);
		else
		{
			bootsect->VolumeSerialNo [0] = (year & 0xFF) + minute;
			bootsect->VolumeSerialNo [1] = (year >> 8)   + hour;
			bootsect->VolumeSerialNo [2] =  day          + msec / 10;
			bootsect->VolumeSerialNo [3] =  month        + second;
		}
	}
	iobuf [SectorSize-2] = 0x55;
	iobuf [SectorSize-1] = 0xAA;
}

void	write_boot	(void)

{
	if (verbosef)
		printf (catgets (catd, NL_SETN, 52, MSG_52), ReservedSectors);
#ifdef USE_SCSI_WRITE
	if (WriteAndVerify (FD, 0UL, ReservedSectors, iobuf) 
		== FAILURE)
	{
		fputs (catgets (catd, NL_SETN, 45, MSG_45), stderr);
		StatusReport ();
	}
#else /* USE_SCSI_WRITE */
	lseek (fd, 0L, SEEK_SET);
	if (write (fd, iobuf, SectorSize * ReservedSectors) <
		SectorSize * ReservedSectors)
		perror (NAME ": write_boot()");
#endif
}

#ifdef DOS_BULK_FAT_WRITE
void	farmemset	(BYTE far *ptr, BYTE chr, WORD paragraph)

{
	WORD	seg, ofs;
	
	seg = FP_SEG(ptr) + (FP_OFF(ptr) >> 4);
	ofs = FP_OFF(ptr) & 15;
	for (; paragraph >= 0x0FFF; paragraph -= 0xFFF)
	{
		memset (MK_FP(seg,ofs), chr, 0xFFF0);
		seg += 0x0FFF;
	}
	memset (MK_FP(seg,ofs), chr, paragraph << 4);
}
#endif

void	write_fat	(void)

{
	int		i;
#if (defined MSDOS && !defined DOS_BULK_FAT_WRITE) || defined freebsd || \
	defined SUN_USE_GD
	int		sect, count, dirsect, fatsect [8];
	BYTE	*f0;
		
	for (i = 0; i < FATS; i++)
		fatsect [i] = ReservedSectors + SectorsPerFAT*i;
	dirsect  = ReservedSectors + SectorsPerFAT * FATS;
		
	if (verbosef)
		printf (catgets (catd, NL_SETN, 54, MSG_54), SectorsPerFAT, FATS);
	memset (iobuf, FAT_FILLER, SectorSize * SectorStep);
	for (sect = fatsect [0]; sect < dirsect; sect += SectorStep)
	{
		count = min (SectorStep, dirsect - sect);
		for (i = 0; i < FATS; i++)
			if (sect <= fatsect [i] && fatsect [i] < sect+count)
			{
				f0 = iobuf + SectorSize * (fatsect [i] - sect);
				f0 [0] = FATID;
				f0 [1] = 0xFF;;
				f0 [2] = 0xFF;
				f0 [3] = 0xFF;
			}
		if (WriteAndVerify (FD, (DWORD) sect, count, iobuf) 
			== FAILURE)
		{
			fputs (catgets (catd, NL_SETN, 45, MSG_45), stderr);
			StatusReport ();
		}
		if (f0 != NULL)
		{
			memset (f0, FAT_FILLER, 4);
			f0 = NULL;
		}
	}
#else
	static char	*default_ordinal [FATS+1] = { "", MSG_55, MSG_56 };
	char	ordinal [20];

#ifdef MSDOS
	farmemset (iobuf, FAT_FILLER, SectorSize / 16 * SectorsPerFAT);
#else
	memset (iobuf, FAT_FILLER, SectorSize * SectorsPerFAT);
#endif
	for (i = 0; i < FATS; i++)
	{
		DWORD	top = ReservedSectors + SectorsPerFAT * i;
			
		iobuf [0] = FATID;
		iobuf [1] = 0xFF;
		iobuf [2] = 0xFF;
		iobuf [3] = 0xFF;
		if (verbosef)
		{
			strcpy (ordinal, 
				catgets (catd, NL_SETN, 55+i, default_ordinal [i+1]));
			printf (catgets (catd, NL_SETN, 57, MSG_57), 
				ordinal, SectorsPerFAT);
		}
#ifdef USE_SCSI_WRITE
		if (WriteAndVerify (FD, top, SectorsPerFAT, iobuf) 
			== FAILURE)
		{
			fputs (catgets (catd, NL_SETN, 45, MSG_45), stderr);
			StatusReport ();
		}
#else /* USE_SCSI_WRITE */
		lseek (fd, SectorSize * top, SEEK_SET);
		if (write (fd, iobuf, SectorSize * SectorsPerFAT) <
			SectorSize * SectorsPerFAT)
			perror (NAME ": write_fat()");
#endif /* USE_SCSI_WRITE */
	}
#endif
}

void	write_root	(void)

{
	int		i;
	DWORD	sect = ReservedSectors + SectorsPerFAT * FATS;
	
	memset (iobuf, DIR_FILLER, SectorSize * RootDirSectors);
	for (i = 0; i < SectorSize * RootDirSectors; i += DIRENT_SIZE)
		iobuf [i] = 0x00;
	if (labelf)
	{
		memset (iobuf, 0x00, DIRENT_SIZE);
		memcpy (iobuf, vlabel, 8 + 3);
		POKEB (&iobuf [11], FA_LABEL);
		POKEW (&iobuf [22], msdostime);
		POKEW (&iobuf [24], msdosdate);
	}
	if (verbosef)
		printf (catgets (catd, NL_SETN, 58, MSG_58), RootDirSectors);
#ifdef USE_SCSI_WRITE
#ifdef SUN_USE_GD
	for (i = RootDirSectors; i > 0; i -= SectorStep, sect += SectorStep)
	{
		if (WriteAndVerify (FD, sect, min (SectorStep, i), iobuf) 
			== FAILURE)
		{
			fputs (catgets (catd, NL_SETN, 45, MSG_45), stderr);
			StatusReport ();
		}
		if (i == RootDirSectors)
			memcpy (iobuf, &iobuf [DIRENT_SIZE], DIRENT_SIZE);
	}
#else
	if (WriteAndVerify (FD, sect, RootDirSectors, iobuf) 
		== FAILURE)
	{
		fputs (catgets (catd, NL_SETN, 45, MSG_45), stderr);
		StatusReport ();
	}
#endif
#else /* USE_SCSI_WRITE */
	lseek (fd, SectorSize * sect, SEEK_SET);
	if (write (fd, iobuf, SectorSize * RootDirSectors) 
		< SectorSize * RootDirSectors)
		perror (NAME ": write_root()");
#endif
}

void	release_iobuf	(void)

{
#ifndef MSDOS
	free (iobuf);
#else
#ifdef DOS_USE_ALLOCMEM
	free (iobuf);
#else
	freemem (iobufseg);
#endif
#endif /* MSDOS */
}

void	close_device	(void)

{
	int		drive;
	
	if (verbosef)
		fputs (catgets (catd, NL_SETN, 59, MSG_59), stdout);
#ifdef MSDOS
	if (ejectf)
	{
		if (verbosef)
			fputs (catgets (catd, NL_SETN, 35, MSG_35), stdout);
		StartStopUnit (FD, 2);
	}
	else if (ispc98 () && 
		(drive = PC98ConvertDAUAIntoDrive (DA_SCSI_CHR + scsi_id)) != FAILURE)
		PC98ForceDPBUpdate (scsi_id, drive);
#else
	CLOSE (fd);
#endif /* MSDOS */
}

void	main	(int argc, char **argv)

{
#ifdef NLS
	setlocale (LC_ALL, "");
	catd = catopen (NAME, CATFLAG);
#endif
	
#ifdef MSDOS
	strcpy (bootfilepath, argv [0]);
#else
	strcpy (bootfilepath, LIBDIR);
#endif
	
	if (verbosef)
		banner ();
	analyze_option (argc, argv);
	if (phyfrmf || boottype != NOBOOT || fatf || rootf)
	{
		open_device ();
		check_medium ();
		if (phyfrmf)
		{
			prompt1 ();
			physical_format ();
		}
		if (boottype != NOBOOT || fatf || rootf)
		{
			setup_time ();
			setup_params ();
			alloc_iobuf ();
			prompt2 ();
			if (boottype != NOBOOT)
			{
				build_boot ();
				write_boot ();
			}
			if (fatf)
				write_fat ();
			if (rootf)
				write_root ();
			release_iobuf ();
		}
		close_device ();
	}
	
#ifdef NLS
	catclose (catd);
#endif
	exit (EXIT_SUCCESS);
}

