/*
	A poor subset implementation on MS-DOS
	of the XPG NLS message catalog system
	Copyright 1995,1996 (C) Ito Takayuki
	
	LARGE DATA MODEL ONLY
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>
#include	<errno.h>
#include	<io.h>
#include	<dos.h>
#include	"typedef.h"
#include	"locale.h"
#include	"msgcat.h"

static char	*lang, *nlspath;
static	BOOL	initf = FALSE;

static struct catd_blk_t
{
	BOOL	openf;
	char	**msgptr;
	char	*msgbuf;
} catd_blk [NL_CATDMAX + 1];

nl_catd	catopen		(const char *name, int oflag)

{
	int		msg_num, set_num, c, l, quotechar;
	size_t	count;
	char	*p, *q, linbuf [NL_TEXTMAX];
	char	catpath [FILENAME_MAX];
	FILE	*fp;
	nl_catd	catd;
	struct catd_blk_t	*blk;
	
	if (oflag != 0 && oflag != NL_CAT_LOCALE)
	{
		errno = EINVAL;
		return FAILURE;
	}
	if (!initf)
	{
		if ((nlspath = getenv ("NLSPATH")) == NULL)
		{
			errno = ENOENT;
			return FAILURE;
		}
		if (oflag == NL_CAT_LOCALE)
			lang = setlocale (LC_MESSAGES, NULL);
		else
			lang = getenv ("LANG");
		if (lang == NULL)
		{
			errno = ENOENT;
			return FAILURE;
		}
		InitDBCSLeadTable ();
		for (catd = NL_CATDMIN; catd <= NL_CATDMAX; catd++)
			catd_blk [catd].openf = FALSE;
		initf = TRUE;
		catd = NL_CATDMIN;
	}
	else
	{
		for (catd = NL_CATDMIN; catd <= NL_CATDMAX; catd++)
			if (catd_blk [catd].openf == FALSE)
				break;
		if (catd > NL_CATDMAX)
		{
			errno = EMFILE;
			return FAILURE;
		}
	}
	
	sprintf (catpath, "%s/%s/%s.cat", nlspath, lang, name);
	if ((fp = fopen (catpath, "rt")) == NULL)
	{
		errno = ENOENT;
		return FAILURE;
	}
	
	blk = &catd_blk [catd];
	if ((blk->msgptr = calloc (NL_MSGMAX, sizeof (char *))) == NULL)
	{
		errno = ENOMEM;
		return FAILURE;
	}
	if ((blk->msgbuf = malloc ((size_t) filelength (fileno (fp)))) == NULL)
	{
		errno = ENOMEM;
		free (blk->msgptr);
		return FAILURE;
	}
	
	q = blk->msgbuf;
	count = 0;
	quotechar = '\0';
	while (fgets (linbuf, NL_TEXTMAX, fp) != NULL)
	{
		l = strlen (linbuf);
		if (linbuf [l-1] == '\n')
			linbuf [--l] = '\0';
		
		if (linbuf [0] == '$')
		{
			if (strncmp (linbuf, "$set", 4) == 0)
			{
				set_num = atoi (&linbuf [5]);
				if (set_num != NL_SETD)
				{
					free (blk->msgptr);
					free (blk->msgbuf);
					errno = E_ILLEGAL_SET;
					return FAILURE;
				}
			}
			else if (strncmp (linbuf, "$quote", 6) == 0)
				quotechar = linbuf [7];
		}
		else
		{
			if (l > MAX_BUFLEN - count)
			{
				free (blk->msgptr);
				free (blk->msgbuf);
				errno = ENOMEM;
				return FAILURE;
			}
			msg_num = atoi (linbuf);
			if (msg_num < NL_MSGMIN || NL_MSGMAX < msg_num)
			{
				free (blk->msgptr);
				free (blk->msgbuf);
				errno = E_ILLEGAL_MSG;
				return FAILURE;
			}
			blk->msgptr [msg_num] = q;
			p = linbuf;
			while (isdigit (*p++));
			if (!isspace (*(p-1)) || 
				(quotechar != '\0' && *p++ != quotechar))
			{
				free (blk->msgptr);
				free (blk->msgbuf);
				errno = E_ILLEGAL_MSG;
				return FAILURE;
			}
			while ((c = *p++) != quotechar)
			{
				if (c == '\\')
				{
					switch (c = *p++)
					{
						case 'a':	*q++ = '\a';	break;
						case 'b':	*q++ = '\b';	break;
						case 'e':	*q++ = '\e';	break;
						case 'n':	*q++ = '\n';	break;
						case 'r':	*q++ = '\r';	break;
						case 't':	*q++ = '\t';	break;
						
						case 'x':
						sscanf (p, "%x", &c);
						*q++ = c;
						while (isxdigit(*p++));
						--p;
						break;
						
						case '\0':
						fgets (linbuf, NL_TEXTMAX, fp);
						l = strlen (linbuf);
						if (linbuf [l-1] == '\n')
							linbuf [--l] = '\0';
						p = linbuf;
						break;
						
						default:
						if (isdigit (c))
						{
							sscanf (--p, "%o", &c);
							*q++ = c;
							while (isdigit (*p++));
							--p;
						}
						else
							*q++ = c;
						break;
					}
				}
				else if (IsDBCSLeadByte ((BYTE)c))
				{
					*q++ = c;
					*q++ = *p++;
				}
				else if (c == '\0')
				{
					fgets (linbuf, NL_TEXTMAX, fp);
					l = strlen (linbuf);
					if (linbuf [l-1] == '\r')
						linbuf [--l] = '\0';
					p = linbuf;
				}
				else
					*q++ = c;
			}
			*q++ = '\0';
			count += strlen (blk->msgptr [msg_num]) + 1;
		}
	}
	
	fclose (fp);
	blk->openf = TRUE;
	return catd;
}

int		catclose	(nl_catd catd)

{
	if (!catd_blk [catd].openf)
	{
		errno = EBADF;
		return SUCCESS;
	}
	else
	{
		free (catd_blk [catd].msgptr);
		free (catd_blk [catd].msgbuf);
		return SUCCESS;
	}
}

char	*catgets
(nl_catd catd, int set_num, int msg_num, const char *def_str)

{
	static char	linbuf [NL_TEXTMAX];
	
	if (catd < NL_CATDMIN || NL_CATDMAX < catd || !catd_blk [catd].openf)
	{
		errno = EBADF;
		return (char *) def_str;
	}
	if (set_num != NL_SETD)
	{
		errno = E_NO_SET;
		return (char *) def_str;
	}
	if (msg_num < NL_MSGMIN || NL_MSGMAX < msg_num || 
		catd_blk [catd].msgptr [msg_num] == NULL)
	{
		errno = E_NO_MSG;
		return (char *) def_str;
	}
	
	strcpy (linbuf, catd_blk [catd].msgptr [msg_num]);
	return linbuf;
}

