/*
	OSF/1
	SCSI device driver interface module
	Copyright 1995,1996 (C) ITO Takayuki
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<fcntl.h>
#include	<sys/ioctl.h>
#include	<sys/types.h>
#include	<netinet/in.h>
#include	<io/common/iotypes.h>
#include	<io/cam/cam.h>
#include	<io/cam/dec_cam.h>
#include	<io/cam/uagt.h>
#include	<io/cam/scsi_all.h>
#include	<io/cam/scsi_direct.h>
#include	<io/cam/scsi_status.h>
#include	"typedef.h"
#include	"osf1scsi.h"

#define	RETRY	5

unsigned	SCSIDriverStatus = CAM_REQ_CMP;
BYTE	SCSIStatus = SCSI_STAT_GOOD;
BYTE	SCSISenseKey = ALL_NO_SENSE;
BYTE	SCSISenseCode = 0;
BYTE	SCSISenseCodeQ = 0;
BOOL	UnitAtnRetry = FALSE;
DWORD	blocksize = 512U;

static int	camfd;

int		OpenCAM	(void)

{
	return camfd = open (DEVICE_FILE, O_RDWR);
}

int		CloseCAM	(void)

{
	return close (camfd);
}

int		RequestSense	(int hbaid, int scsiid, int lun, BYTE *buf, int buflen)

{
	int		stat;
	static UAGT_CAM_CCB	ua_ccb;
	static CCB_SCSIIO	ccb;
	ALL_REQ_SENSE_CDB6	*cdb;
	
	memset (&ccb, 0, sizeof ccb);
	ccb.cam_ch.my_addr = (struct ccb_header *)&ccb;
	ccb.cam_ch.cam_ccb_len = sizeof ccb;
	ccb.cam_ch.cam_func_code = XPT_SCSI_IO;
	ccb.cam_ch.cam_path_id = hbaid;
	ccb.cam_ch.cam_target_id = scsiid;
	ccb.cam_ch.cam_target_lun = lun;
	ccb.cam_ch.cam_flags = CAM_DIR_IN;
	
	ccb.cam_data_ptr = buf;
	ccb.cam_dxfer_len = buflen;
	ccb.cam_timeout = CAM_TIME_DEFAULT;
	ccb.cam_cdb_len = sizeof (ALL_REQ_SENSE_CDB6);
	
	cdb = (ALL_REQ_SENSE_CDB6 *) ccb.cam_cdb_io.cam_cdb_bytes;
	cdb->opcode = ALL_REQ_SENSE6_OP;
	cdb->alloc_len = buflen;
	
	memset (&ua_ccb, 0, sizeof ua_ccb);
	ua_ccb.uagt_ccb = (CCB_HEADER *) &ccb;
	ua_ccb.uagt_ccblen = sizeof ccb;
	ua_ccb.uagt_buffer = buf;
	ua_ccb.uagt_buflen = buflen;
	
	/* CAMPerform () */
	ua_ccb.uagt_cdb = NULL;
	ua_ccb.uagt_cdblen = 0;
	ua_ccb.uagt_ccb->cam_flags |= CAM_DIS_AUTOSENSE;
	ua_ccb.uagt_snsbuf = NULL;
	ua_ccb.uagt_snslen = 0;
	
	if (ioctl (camfd, UAGT_CAM_IO, &ua_ccb) == FAILURE)
		stat = FAILURE;
	else if ((SCSIDriverStatus = ua_ccb.uagt_ccb->cam_status) != CAM_REQ_CMP)
		stat = FAILURE;
	else if ((SCSIStatus = ((CCB_SCSIIO *) ua_ccb.uagt_ccb)->cam_scsi_status) 
		== SCSI_STAT_GOOD)
	{
		SCSISenseKey   = ((ALL_REQ_SNS_DATA *) buf)->sns_key;
		SCSISenseCode  = ((ALL_REQ_SNS_DATA *) buf)->asc;
		SCSISenseCodeQ = ((ALL_REQ_SNS_DATA *) buf)->asq;
		stat = SUCCESS;
	}
	else
	{
		if (SCSIStatus == SCSI_STAT_CHECK_CONDITION)
			RequestSense (hbaid, scsiid, lun, buf, buflen);
		stat = FAILURE;
	}
	return stat;
}

static int	CAMPerform	(UAGT_CAM_CCB *ua_ccb, int retry)

{
	int		stat;
	ALL_REQ_SNS_DATA	sense_data;
	UAGT_CAM_CCB	ua_ccb_sim_rel, ua_ccb_reset_dev;
	CCB_RELSIM		ccb_sim_rel;
	CCB_RESETDEV	ccb_reset_dev;

	if (retry == 0)
		return FAILURE;
	
	fputs ("CAMPerform (): enter\n", stderr);
	ua_ccb->uagt_cdb = NULL;
	ua_ccb->uagt_cdblen = 0;
	ua_ccb->uagt_ccb->cam_flags |= CAM_DIS_AUTOSENSE;
	ua_ccb->uagt_snsbuf = NULL;
	ua_ccb->uagt_snslen = 0;
	
	stat = ioctl (camfd, UAGT_CAM_IO, ua_ccb);
	fputs ("CAMPerform (): 0\n", stderr);
	if (stat == FAILURE)
	{
		fputs ("CAMPerform (): 10\n", stderr);
		stat = FAILURE;
	}
	else if (((SCSIDriverStatus = ua_ccb->uagt_ccb->cam_status) 
		& CAM_STATUS_MASK) != CAM_REQ_CMP)
	{
		fputs ("CAMPerform (): 20\n", stderr);
		stat = FAILURE;
		if (SCSIDriverStatus & CAM_SIM_QFRZN)
		{
			ccb_sim_rel.cam_ch.my_addr = (struct ccb_header *)&ccb_sim_rel;
			ccb_sim_rel.cam_ch.cam_ccb_len = sizeof ccb_sim_rel;
			ccb_sim_rel.cam_ch.cam_func_code = XPT_REL_SIMQ;
			ccb_sim_rel.cam_ch.cam_path_id   =ua_ccb->uagt_ccb->cam_path_id;
			ccb_sim_rel.cam_ch.cam_target_id =ua_ccb->uagt_ccb->cam_target_id;
			ccb_sim_rel.cam_ch.cam_target_lun=ua_ccb->uagt_ccb->cam_target_lun;
			ccb_sim_rel.cam_ch.cam_flags = CAM_DIR_NONE;
			memset (&ua_ccb_sim_rel, 0, sizeof ua_ccb_sim_rel);
			ua_ccb_sim_rel.uagt_ccb = (CCB_HEADER *)&ccb_sim_rel;
			ua_ccb_sim_rel.uagt_ccblen = sizeof ccb_sim_rel;
			if (ioctl (camfd, UAGT_CAM_IO, &ua_ccb_sim_rel) == SUCCESS &&
				(ccb_sim_rel.cam_ch.cam_status & CAM_STATUS_MASK) == CAM_REQ_CMP)
			{
#if 0
				ccb_reset_dev.cam_ch.my_addr = (struct ccb_header *)&ccb_reset_dev;
				ccb_reset_dev.cam_ch.cam_ccb_len = sizeof ccb_reset_dev;
				ccb_reset_dev.cam_ch.cam_func_code = XPT_RESET_DEV;
				ccb_reset_dev.cam_ch.cam_path_id   =ua_ccb->uagt_ccb->cam_path_id;
				ccb_reset_dev.cam_ch.cam_target_id =ua_ccb->uagt_ccb->cam_target_id;
				ccb_reset_dev.cam_ch.cam_target_lun=ua_ccb->uagt_ccb->cam_target_lun;
				ccb_reset_dev.cam_ch.cam_flags = CAM_DIR_NONE;
				memset (&ua_ccb_reset_dev, 0, sizeof ua_ccb_reset_dev);
				ua_ccb_reset_dev.uagt_ccb = (CCB_HEADER *)&ccb_reset_dev;
				ua_ccb_reset_dev.uagt_ccblen = sizeof ccb_reset_dev;
				if (ioctl (camfd, UAGT_CAM_IO, &ua_ccb_reset_dev) == SUCCESS &&
					(ccb_reset_dev.cam_ch.cam_status & CAM_STATUS_MASK) == CAM_REQ_CMP);
#endif
			}
			RequestSense (
				ua_ccb->uagt_ccb->cam_path_id,
				ua_ccb->uagt_ccb->cam_target_id,
				ua_ccb->uagt_ccb->cam_target_lun,
				&sense_data, sizeof sense_data);
			stat = CAMPerform (ua_ccb, retry - 1);
		}
	}
	else if ((SCSIStatus = ((CCB_SCSIIO *) ua_ccb->uagt_ccb)->cam_scsi_status) 
		== SCSI_STAT_GOOD)
	{
		fputs ("CAMPerform (): 30\n", stderr);
		stat = SUCCESS;
	}
	else
	{
		stat = FAILURE;
		if (SCSIStatus == SCSI_STAT_CHECK_CONDITION)
		{
			fputs ("CAMPerform (): 40\n", stderr);
			RequestSense (
				ua_ccb->uagt_ccb->cam_path_id,
				ua_ccb->uagt_ccb->cam_target_id,
				ua_ccb->uagt_ccb->cam_target_lun,
				&sense_data, sizeof sense_data);
			if (SCSISenseKey == ALL_UNIT_ATTEN && UnitAtnRetry)
			{
				fputs ("CAMPerform (): 45\n", stderr);
				stat = CAMPerform (ua_ccb, retry - 1);
			}
		}
	}
	return stat;
}

static int	_ReadWrite10	(int hbaid, int scsiid, int lun,
							DWORD lba, WORD count, void *buf, int func)

{
	static UAGT_CAM_CCB	ua_ccb;
	static CCB_SCSIIO	ccb;
	DIR_READ_CDB10	*cdb;
	
	memset (&ccb, 0, sizeof ccb);
	ccb.cam_ch.my_addr = (struct ccb_header *)&ccb;
	ccb.cam_ch.cam_ccb_len = sizeof ccb;
	ccb.cam_ch.cam_func_code = XPT_SCSI_IO;
	ccb.cam_ch.cam_path_id = hbaid;
	ccb.cam_ch.cam_target_id = scsiid;
	ccb.cam_ch.cam_target_lun = lun;
	ccb.cam_ch.cam_flags = (func ? CAM_DIR_IN : CAM_DIR_OUT);
	
	ccb.cam_data_ptr = buf;
	ccb.cam_dxfer_len = blocksize * (func == 3 ? 1 : count);
	ccb.cam_timeout = CAM_TIME_DEFAULT;
	ccb.cam_cdb_len = sizeof (DIR_READ_CDB10);
	
	cdb = (DIR_READ_CDB10 *) ccb.cam_cdb_io.cam_cdb_bytes;
	switch (func)
	{
		case 0:	cdb->opcode = DIR_READ10_OP;		break;
		case 1:	cdb->opcode = DIR_WRITE10_OP;		break;
		case 2:	cdb->opcode = DIR_WRITE_VRFY_10_OP;	break;
		case 3:	cdb->opcode = 0x41;					break;
	}
	DIRLBN_TO_READ10 (lba,cdb);
	DIRTRANS_TO_READ10 (count,cdb);
	
	memset (&ua_ccb, 0, sizeof ua_ccb);
	ua_ccb.uagt_ccb = (CCB_HEADER *) &ccb;
	ua_ccb.uagt_ccblen = sizeof ccb;
	ua_ccb.uagt_buffer = buf;
	ua_ccb.uagt_buflen = blocksize * (func == 3 ? 1 : count);
	
	return CAMPerform (&ua_ccb, RETRY);
}

int	Read10			(int hbaid, int scsiid, int lun,
					DWORD lba, WORD count, void *buf)

{
	return _ReadWrite10 (hbaid, scsiid, lun, lba, count, buf, 0);
}

int	Write10			(int hbaid, int scsiid, int lun,
					DWORD lba, WORD count, void *buf)

{
	return _ReadWrite10 (hbaid, scsiid, lun, lba, count, buf, 1);
}

int	WriteAndVerify	(int hbaid, int scsiid, int lun,
					DWORD lba, WORD count, void *buf)

{
	return _ReadWrite10 (hbaid, scsiid, lun, lba, count, buf, 2);
}

int	WriteSame		(int hbaid, int scsiid, int lun,
					DWORD lba, WORD count, void *buf)

{
	return _ReadWrite10 (hbaid, scsiid, lun, lba, count, buf, 3);
}

int		FormatUnit	(int hbaid, int scsiid, int lun)

{
	static UAGT_CAM_CCB	ua_ccb;
	static CCB_SCSIIO	ccb;
	DIR_FORMAT_CDB6	*cdb;
	
	memset (&ccb, 0, sizeof ccb);
	ccb.cam_ch.my_addr = (struct ccb_header *)&ccb;
	ccb.cam_ch.cam_ccb_len = sizeof ccb;
	ccb.cam_ch.cam_func_code = XPT_SCSI_IO;
	ccb.cam_ch.cam_path_id = hbaid;
	ccb.cam_ch.cam_target_id = scsiid;
	ccb.cam_ch.cam_target_lun = lun;
	ccb.cam_ch.cam_flags = 0;
	
	ccb.cam_data_ptr = NULL;
	ccb.cam_dxfer_len = 0;
	ccb.cam_timeout = CAM_TIME_DEFAULT;
	ccb.cam_cdb_len = sizeof (DIR_FORMAT_CDB6);
	
	cdb = (DIR_FORMAT_CDB6 *) ccb.cam_cdb_io.cam_cdb_bytes;
	cdb->opcode = DIR_FORMAT_OP;
	
	memset (&ua_ccb, 0, sizeof ua_ccb);
	ua_ccb.uagt_ccb = (CCB_HEADER *) &ccb;
	ua_ccb.uagt_ccblen = sizeof ccb;
	ua_ccb.uagt_buffer = NULL;
	ua_ccb.uagt_buflen = 0;

	return CAMPerform (&ua_ccb, RETRY);
}

int		TestUnitReady	(int hbaid, int scsiid, int lun)

{
	static UAGT_CAM_CCB	ua_ccb;
	static CCB_SCSIIO	ccb;
	ALL_TUR_CDB	*cdb;
	
	memset (&ccb, 0, sizeof ccb);
	ccb.cam_ch.my_addr = (struct ccb_header *)&ccb;
	ccb.cam_ch.cam_ccb_len = sizeof ccb;
	ccb.cam_ch.cam_func_code = XPT_SCSI_IO;
	ccb.cam_ch.cam_path_id = hbaid;
	ccb.cam_ch.cam_target_id = scsiid;
	ccb.cam_ch.cam_target_lun = lun;
	ccb.cam_ch.cam_flags = 0;
	
	ccb.cam_data_ptr = NULL;
	ccb.cam_dxfer_len = 0;
	ccb.cam_timeout = CAM_TIME_DEFAULT;
	ccb.cam_cdb_len = sizeof (ALL_TUR_CDB);
	
	cdb = (ALL_TUR_CDB *) ccb.cam_cdb_io.cam_cdb_bytes;
	cdb->opcode = ALL_TUR_OP;
	
	memset (&ua_ccb, 0, sizeof ua_ccb);
	ua_ccb.uagt_ccb = (CCB_HEADER *) &ccb;
	ua_ccb.uagt_ccblen = sizeof ccb;
	ua_ccb.uagt_buffer = NULL;
	ua_ccb.uagt_buflen = 0;

	return CAMPerform (&ua_ccb, RETRY);
}

int		StartStopUnit	(int hbaid, int scsiid, int lun, int flag)

{
	static UAGT_CAM_CCB	ua_ccb;
	static CCB_SCSIIO	ccb;
	DIR_START_CDB6	*cdb;
	
	memset (&ccb, 0, sizeof ccb);
	ccb.cam_ch.my_addr = (struct ccb_header *)&ccb;
	ccb.cam_ch.cam_ccb_len = sizeof ccb;
	ccb.cam_ch.cam_func_code = XPT_SCSI_IO;
	ccb.cam_ch.cam_path_id = hbaid;
	ccb.cam_ch.cam_target_id = scsiid;
	ccb.cam_ch.cam_target_lun = lun;
	ccb.cam_ch.cam_flags = 0;
	
	ccb.cam_data_ptr = NULL;
	ccb.cam_dxfer_len = 0;
	ccb.cam_timeout = CAM_TIME_DEFAULT;
	ccb.cam_cdb_len = sizeof (DIR_START_CDB6);
	
	cdb = (DIR_START_CDB6 *) ccb.cam_cdb_io.cam_cdb_bytes;
	cdb->opcode = DIR_START_OP;
	cdb->start = flag & 1;
	cdb->loej  = (flag >> 1) & 1;
	
	memset (&ua_ccb, 0, sizeof ua_ccb);
	ua_ccb.uagt_ccb = (CCB_HEADER *) &ccb;
	ua_ccb.uagt_ccblen = sizeof ccb;
	ua_ccb.uagt_buffer = NULL;
	ua_ccb.uagt_buflen = 0;

	return CAMPerform (&ua_ccb, RETRY);
}

int		PreventAllowMediumRemoval	(int hbaid, int scsiid, int lun, int flag)

{
	static UAGT_CAM_CCB	ua_ccb;
	static CCB_SCSIIO	ccb;
	DIR_PREVENT_CDB6	*cdb;
	
	memset (&ccb, 0, sizeof ccb);
	ccb.cam_ch.my_addr = (struct ccb_header *)&ccb;
	ccb.cam_ch.cam_ccb_len = sizeof ccb;
	ccb.cam_ch.cam_func_code = XPT_SCSI_IO;
	ccb.cam_ch.cam_path_id = hbaid;
	ccb.cam_ch.cam_target_id = scsiid;
	ccb.cam_ch.cam_target_lun = lun;
	ccb.cam_ch.cam_flags = 0;
	
	ccb.cam_data_ptr = NULL;
	ccb.cam_dxfer_len = 0;
	ccb.cam_timeout = CAM_TIME_DEFAULT;
	ccb.cam_cdb_len = sizeof (DIR_PREVENT_CDB6);
	
	cdb = (DIR_PREVENT_CDB6 *) ccb.cam_cdb_io.cam_cdb_bytes;
	cdb->opcode = DIR_PREVENT_OP;
	cdb->prevent = flag;
	
	memset (&ua_ccb, 0, sizeof ua_ccb);
	ua_ccb.uagt_ccb = (CCB_HEADER *) &ccb;
	ua_ccb.uagt_ccblen = sizeof ccb;
	ua_ccb.uagt_buffer = NULL;
	ua_ccb.uagt_buflen = 0;

	return CAMPerform (&ua_ccb, RETRY);
}

int		ReadCapacity	(int hbaid, int scsiid, int lun, DWORD *buf)

{
	static UAGT_CAM_CCB	ua_ccb;
	static CCB_SCSIIO	ccb;
	DIR_READ_CAP_CDB10	*cdb;
	int	result;
	
	memset (&ccb, 0, sizeof ccb);
	ccb.cam_ch.my_addr = (struct ccb_header *)&ccb;
	ccb.cam_ch.cam_ccb_len = sizeof ccb;
	ccb.cam_ch.cam_func_code = XPT_SCSI_IO;
	ccb.cam_ch.cam_path_id = hbaid;
	ccb.cam_ch.cam_target_id = scsiid;
	ccb.cam_ch.cam_target_lun = lun;
	ccb.cam_ch.cam_flags = CAM_DIR_IN;
	
	ccb.cam_data_ptr = buf;
	ccb.cam_dxfer_len = 8;
	ccb.cam_timeout = CAM_TIME_DEFAULT;
	ccb.cam_cdb_len = sizeof (DIR_READ_CAP_CDB10);
	
	cdb = (DIR_READ_CAP_CDB10 *) ccb.cam_cdb_io.cam_cdb_bytes;
	cdb->opcode = DIR_READCAP_OP;
	
	memset (&ua_ccb, 0, sizeof ua_ccb);
	ua_ccb.uagt_ccb = (CCB_HEADER *) &ccb;
	ua_ccb.uagt_ccblen = sizeof ccb;
	ua_ccb.uagt_buffer = buf;
	ua_ccb.uagt_buflen = 8;

	if ((result = CAMPerform (&ua_ccb, RETRY)) == SUCCESS)
	{
		buf [0] = ntohl (buf [0]);
		buf [1] = ntohl (buf [1]);
	}
	return result;
}

int		ModeSense	(int hbaid, int scsiid, int lun, BYTE *buf, int buflen)

{
	static UAGT_CAM_CCB	ua_ccb;
	static CCB_SCSIIO	ccb;
	ALL_MODE_SENSE_CDB6	*cdb;
	
	if  (buflen > 255) buflen = 255;
	memset (&ccb, 0, sizeof ccb);
	ccb.cam_ch.my_addr = (struct ccb_header *)&ccb;
	ccb.cam_ch.cam_ccb_len = sizeof ccb;
	ccb.cam_ch.cam_func_code = XPT_SCSI_IO;
	ccb.cam_ch.cam_path_id = hbaid;
	ccb.cam_ch.cam_target_id = scsiid;
	ccb.cam_ch.cam_target_lun = lun;
	ccb.cam_ch.cam_flags = CAM_DIR_IN;
	
	ccb.cam_data_ptr = buf;
	ccb.cam_dxfer_len = buflen;
	ccb.cam_timeout = CAM_TIME_DEFAULT;
	ccb.cam_cdb_len = sizeof (ALL_MODE_SENSE_CDB6);
	
	cdb = (ALL_MODE_SENSE_CDB6 *) ccb.cam_cdb_io.cam_cdb_bytes;
	cdb->opcode = ALL_MODE_SENSE6_OP;
	cdb->alloc_len = buflen;
	
	memset (&ua_ccb, 0, sizeof ua_ccb);
	ua_ccb.uagt_ccb = (CCB_HEADER *) &ccb;
	ua_ccb.uagt_ccblen = sizeof ccb;
	ua_ccb.uagt_buffer = buf;
	ua_ccb.uagt_buflen = buflen;

	return CAMPerform (&ua_ccb, RETRY);
}

int		Inquiry	(int hbaid, int scsiid, int lun, BYTE *buf, int buflen)

{
	static UAGT_CAM_CCB	ua_ccb;
	static CCB_SCSIIO	ccb;
	ALL_INQ_CDB	*cdb;
	
	if  (buflen > 255) buflen = 255;
	memset (&ccb, 0, sizeof ccb);
	ccb.cam_ch.my_addr = (struct ccb_header *)&ccb;
	ccb.cam_ch.cam_ccb_len = sizeof ccb;
	ccb.cam_ch.cam_func_code = XPT_SCSI_IO;
	ccb.cam_ch.cam_path_id = hbaid;
	ccb.cam_ch.cam_target_id = scsiid;
	ccb.cam_ch.cam_target_lun = lun;
	ccb.cam_ch.cam_flags = CAM_DIR_IN;
	
	ccb.cam_data_ptr = buf;
	ccb.cam_dxfer_len = buflen;
	ccb.cam_timeout = CAM_TIME_DEFAULT;
	ccb.cam_cdb_len = sizeof (ALL_INQ_CDB);
	
	cdb = (ALL_INQ_CDB *) ccb.cam_cdb_io.cam_cdb_bytes;
	cdb->opcode = ALL_INQ_OP;
	cdb->alloc_len = buflen;
	
	memset (&ua_ccb, 0, sizeof ua_ccb);
	ua_ccb.uagt_ccb = (CCB_HEADER *) &ccb;
	ua_ccb.uagt_ccblen = sizeof ccb;
	ua_ccb.uagt_buffer = buf;
	ua_ccb.uagt_buflen = buflen;

	return CAMPerform (&ua_ccb, RETRY);
}

