/*
	setlocale(3) for MS-DOS
	Copyright 1996 (C) ITO Takayuki
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<io.h>
#include	<fcntl.h>
#include	<dos.h>
#include	"typedef.h"
#include	"locale.h"

#undef	DEBUG

/*
	locale category names and locale names
*/

struct
{
	char	*envvarname;
	char	locname [NL_LANGMAX];
} LocaleDB [] =
{
	{ "LC_COLLATE",	C_LOCALE },
	{ "LC_CTYPE",	C_LOCALE },
	{ "LC_MONETARY",C_LOCALE },
	{ "LC_NUMERIC",	C_LOCALE },
	{ "LC_TIME",	C_LOCALE },
	{ "LC_MESSAGES",C_LOCALE }
};

/* %NLSPATH% value */

static char	*nlspath = NULL;

/* DBCS lead-in byte table: array of boolean */

BYTE	DBCSLeadTable [0x100];

/* DOS function 6300h: get DBCS vector */

BYTE	far *DOSGetDBCSVector	(void)

{
	union REGS	regs;
	struct SREGS	sregs;
	
	regs.x.ax = 0x6300;
	segread (&sregs);
	intdosx (&regs, &regs, &sregs);
	if (regs.h.al == 0xFF)
		return (BYTE far *) NULL;
	return MK_FP (sregs.ds, regs.x.si);
}

/* initialize DBCS lead-in byte table */

void	InitDBCSLeadTable	(void)

{
	int		i, end;
	BYTE	far *DBCSVector;
	
	memset (DBCSLeadTable, FALSE, sizeof DBCSLeadTable);
	if ((DBCSVector = DOSGetDBCSVector ()) != NULL)
		while (i = *DBCSVector++, end = *DBCSVector++, 
			(i | end) != 0)
		{
			for ( ; i <= end ; i++)
				DBCSLeadTable [i] = TRUE;
		}
	return;
}

/* DOS function 2F/AD80h: get current keyboard code */

char	far *DOSGetKbdCode	(void)

{
	union REGS	regs;
	struct SREGS	sregs;
	
	regs.x.ax = 0xAD80;
	segread (&sregs);
	int86x (0x2F, &regs, &regs, &sregs);
	if (regs.h.al != 0xFF)
		return (char far *) NULL;
	return MK_FP (sregs.es, regs.x.di + 0x16);
}

/* DOS function 6501h: get current country code and code page # */

int		DOSGetCurrentCountry	(WORD *country, WORD *codepage)

{
	BYTE	buf [7];
	union REGS		regs;
	struct SREGS	sregs;
	
	regs.x.ax = 0x6501;
	regs.x.bx = regs.x.dx = 0xFFFF;
	regs.x.cx = sizeof buf;
	regs.x.di = FP_OFF(buf);
	sregs.es  = FP_SEG(buf);
	intdosx (&regs, &regs, &sregs);
	if (regs.x.cflag)
	{
#ifdef __TURBOC__
		_doserrno = regs.x.ax;
#endif
		return FAILURE;
	}
	*country  = *(WORD *) &buf [3];
	*codepage = *(WORD *) &buf [5];
	return SUCCESS;
}

static struct aliasdir_t	*aliasdirbuf = NULL;
static int	aliases = 0;

/*
	load locale alias file into dynamically allocated buffer
	return: SUCCESS/FAILURE
*/

static int	LoadLocaleAlias	(void)

{
	int		result = FAILURE, fd;
	WORD	len;
	char	locpath [FILENAME_MAX];
	
	if (nlspath == NULL)
		return FAILURE;
	sprintf (locpath, "%s/" ALIASFILE, nlspath);
	if ((fd = open (locpath, O_RDONLY | O_BINARY)) != FAILURE)
	{
		len = (WORD) filelength (fd);
		if ((aliasdirbuf = malloc (len)) != NULL &&
			read (fd, (void *) aliasdirbuf, len) == len)
		{
			aliases = len / sizeof (struct aliasdir_t);
			result = SUCCESS;
		}
		close (fd);
	}
	return result;
}

/*
	get canonical locale name for given locale name
	return: name or NULL (load failure or no canonical name)
*/

static char	*ResolveLocaleAlias	(char *alias)

{
	char	*result = NULL;
	int		i;
	
	if (aliases == 0 && LoadLocaleAlias () == FAILURE)
		return NULL;
	for (i = 0; i < aliases; i++)
	{
		if (strcmp (alias, aliasdirbuf [i].alias) == 0 ||
			strcmp (alias, aliasdirbuf [i].locname) == 0)
		{
			result = aliasdirbuf [i].locname;
			break;
		}
	}
	return result;
}

static struct localedir_t	*localedirbuf = NULL;
static int	locales = 0;

/*
	load locale definition file
	return: SUCCESS/FAILURE
*/

static int	LoadLocaleDir	(void)

{
	int		result = FAILURE, fd;
	WORD	len;
	char	locpath [FILENAME_MAX];
	
	if (nlspath == NULL)
		return FAILURE;
	sprintf (locpath, "%s/" LOCALEFILE, nlspath);
	if ((fd = open (locpath, O_RDONLY | O_BINARY)) != FAILURE)
	{
		len = (WORD) filelength (fd);
		if ((localedirbuf = malloc (len)) != NULL &&
			read (fd, (void *) localedirbuf, len) == len)
		{
			locales = len / sizeof (struct localedir_t);
			result = SUCCESS;
		}
		close (fd);
	}
	return result;
}

/*
	detect implicit locale from DOS country code etc.
	return: name or NULL (load failure or unknown locale)
*/

char	*DetectDOSLocale	(void)

{
	WORD	country, codepage, kbdcode, cc, cp, kb;
	char	*loc = NULL, far *p;
	int		i;
	
	if (locales == 0 && LoadLocaleDir () == FAILURE)
		return NULL;

	if (DOSGetCurrentCountry (&country, &codepage) == SUCCESS)
	{
		p = DOSGetKbdCode ();
		kbdcode = (p == NULL ? 0 : *(WORD far *) p);

		for (i = 0; i < locales; i++)
		{
			cc = localedirbuf [i].country;
			cp = localedirbuf [i].codepage;
			kb = localedirbuf [i].kbdcode;
			if (cc == country && 
				(cp == 0 || cp == codepage) &&
				(kb == 0 || kbdcode == 0 || kb == kbdcode))
			{
				/* known locale */
				loc = localedirbuf [i].locname;
				break;
			}
		}
	}
	
	return loc;
}

/*
	set locale name for a single category

	if locale is a null string ("") and any environment variable gives locale,
		the canonical name of the locale is adopted.
	if locale is an explicit locale name,
		the canonical name of the locale is adopted.
	if locale is a NULL pointer or no environment variable gives locale,
		the locale is determined by the DOS country code.

	selected locale (may be overwritten later) is returned.
	if locale is invalid, nothing is done and NULL is returned.
*/

static const char	*SetSingleLocale	(int category, const char *locale)

{
	char	*envval;
	
	/* get locale name from env. var. if locale == "" */
	if (locale != NULL && locale [0] == '\0')
	{
		/* null string: set by env */
		envval = getenv ("LC_ALL");
		if (envval != NULL && envval [0] != '\0')
			/* if %LC_ALL% is valid, it is adopted */
			locale = envval;
		else
		{
			envval = getenv (LocaleDB [category - LC_MIN].envvarname);
			if (envval != NULL && envval [0] != '\0')
				/* if corresponding env. var. is valid, it is adopted */
				locale = envval;
			else
			{
				/* LANG should give default locale */
				envval = getenv ("LANG");
				if (envval != NULL && envval [0] != '\0')
					locale = envval;
				else
					/* no env. var. specifies locale explicitly */
					locale = NULL;
			}
		}
	}

	/* determine canonical locale name */
	if (locale != NULL)
	{
		/* if locale is given explicitly or obtained from env. var. */
		locale = ResolveLocaleAlias ((char *) locale);
#ifdef DEBUG
		fprintf (stderr, "ResolveLocaleAlias() = %s\n", locale);
#endif
	}
	else
	{
		/* if locale is NULL or no env. var. gives it */
		locale = DetectDOSLocale ();
#ifdef DEBUG
		fprintf (stderr, "DetectDOSLocale () = %s\n", locale);
#endif
	}

	/* set category to locale */
	if (locale != NULL)
		strcpy (LocaleDB [category - LC_MIN].locname, locale);
	return locale;
}

#define	LOCALE_DELIM	'/'

/*
	setlocale(3) for DOS

	synopsis and description:

	setlocale (LC_xxx, "");
		set categorie(s) to locale given by env. var.
		(or DOS country code if none)

	setlocale (LC_xxx, "locale_name");
		set categorie(s) to given locale

	setlocale (LC_ALL, "/locale_name/locale_name/...");
		set some cagegories (from LC_COLLATE) to given locales

	setlocale (LC_xxx, NULL);
		query locale(s): no categories modified

	return values:

	for LC_ALL:
	pointer to concatenated locale names for all categories
	"/locale_name/locale_name/..../locale_name"

	for single category:
	pointer to locale name for the category
	or NULL if locale is invalid

	if %NLSPATH% is not set:
	NULL
*/

char	*setlocale	(int category, const char *locale)

{
	char	*p, *result;
	static char	all_lc_buf [NL_LANGMAX*(LC_MAX-LC_MIN+1)+1];
	int		_category;
	
	/* check %NLSPATH% presence */
	if (nlspath == NULL && (nlspath = getenv ("NLSPATH")) == NULL)
		return NULL;

	/* set locale(s) */
	if (locale != NULL)		/* if NULL, only query */
	{
		if (category == LC_ALL)
		{
			/* set locale for all categories */
			if (locale [0] == LOCALE_DELIM)
			{
				/* some locale names for each category */
				locale++;
				for (_category = LC_MIN; _category <= LC_MAX; _category++)
				{
					/* replace next LOCALE_DELIM with \0 */
					if ((p = strchr (locale, LOCALE_DELIM)) != NULL)
						*p = '\0';
					SetSingleLocale (_category, locale);
					/* if there's next name, advance */
					if (p != NULL)
						locale = p+1;
				}
			}
			else
				/* one locale name for all categories */
				for (_category = LC_MIN; _category <= LC_MAX; _category++)
					SetSingleLocale (_category, locale);
		}
		else
			/* set locale for single category */
			result = (char *) SetSingleLocale (category, locale);
	}

	/* return result */
	if (category == LC_ALL)
	{
		/* return concatenated locale names for each category */
		p = all_lc_buf;
		for (_category = LC_MIN; _category <= LC_MAX; _category++)
		{
			*p++ = LOCALE_DELIM;
#ifdef NO_STPCPY
			strcpy (p, LocaleDB [_category - LC_MIN].locname);
			p += strlen (p);
#else
			p = stpcpy (p, LocaleDB [_category - LC_MIN].locname);
#endif
		}
		result = all_lc_buf;
	}
	else
		/* return locale name for the category */
		result = LocaleDB [category - LC_MIN].locname;
	return result;
}

#ifndef __TURBOC__
/* get DOS country info */

struct COUNTRY	*country (int ctcode, struct COUNTRY *ctinfo)

{
	union REGS		regs;
	struct SREGS	sregs;
	
	regs.h.ah = 0x38;
	if (ctcode < 0xFF)
		regs.h.al = ctcode;
	else
	{
		regs.h.al = 0xFF;
		regs.x.bx = ctcode;
	}
	if ((int) ctinfo == -1)
		regs.x.dx = 0xFFFF;
	else
	{
		regs.x.dx = FP_OFF (ctinfo);
		if ((sregs.ds = FP_SEG (ctinfo)) == 0)
			segread (&sregs);
	}
	intdosx (&regs, &regs, &sregs);
	if (regs.x.cflag)
	{
		errno = regs.x.ax;
		return NULL;
	}
	return ctinfo;
}
#endif

/* localeconv(3) */

struct lconv	*localeconv	(void)

{
	static struct COUNTRY	ctinfo;
	static struct lconv		lconv;
	
	if (country (0, &ctinfo) == NULL)
		return NULL;
	lconv.decimal_point = ctinfo.co_desep;
	lconv.thousands_sep = ctinfo.co_thsep;
	lconv.grouping = "\3";							/* fixed */
	lconv.int_curr_symbol = "???";					/* not available */
	lconv.currency_symbol = ctinfo.co_curr;
	lconv.mon_decimal_point = 
		((ctinfo.co_currstyle & 4) ? 				/* use currency symbol */
		ctinfo.co_curr : 							/* as monetary decimal */
		ctinfo.co_desep);							/* same as numeric */
	lconv.mon_thousands_sep = ctinfo.co_thsep;		/* same as numeric */
	lconv.mon_grouping = "\3";						/* fixed */
	lconv.positive_sign = "";
	lconv.negative_sign = "-";
	lconv.int_frac_digits = ctinfo.co_digits;
	lconv.    frac_digits = ctinfo.co_digits;
	lconv.p_cs_precedes  = 
	lconv.n_cs_precedes  = ~(ctinfo.co_currstyle & 1);
	lconv.p_sep_by_space = 
	lconv.n_sep_by_space =  (ctinfo.co_currstyle >> 1) & 1;
	lconv.p_sign_posn = 1;							/* precedes currency */
	lconv.n_sign_posn = 1;							/* precedes currency */
	return &lconv;
}

