/*
	SunOS 4.x / 5.x
	SCSI device driver interface module
	Copyright 1995,1996 (C) ITO Takayuki
*/

#include	<sys/types.h>
#include	<string.h>
#include	<sys/ioctl.h>
#include	<netinet/in.h>
#ifdef sunos4
	#include	<scsi/scsi.h>
	#include	<scsi/impl/uscsi.h>
#else /* sunos5 */
	#include	<sys/scsi/scsi.h>
	#include	<sys/scsi/impl/uscsi.h>
#endif
#include	"typedef.h"
#include	"sunscsi.h"

unsigned	SCSIDriverStatus = 0;
BYTE	SCSIStatus = STATUS_GOOD;
BYTE	SCSISenseKey = 0;
BYTE	SCSISenseCode = 0;
BYTE	SCSISenseCodeQ = 0;
BOOL	UnitAtnRetry = TRUE;
DWORD	blocksize = 512U;

int		RequestSense	(int fd, BYTE *buf, int buflen)

{
	int		stat;
	static struct uscsi_cmd	usc;
	static BYTE	cdb [6];

	if (buflen > 255) buflen = 255;
	memset (cdb, 0, sizeof cdb);
	memset (&usc, 0, sizeof usc);
	cdb [0] = SCMD_REQUEST_SENSE;
	usc.uscsi_cdb = cdb;
	usc.uscsi_cdblen = sizeof cdb;
	usc.uscsi_bufaddr = buf;
	usc.uscsi_buflen = buflen;
	usc.uscsi_flags = USCSI_DIAGNOSE | USCSI_ISOLATE | USCSI_READ;
	
#ifdef sunos5
	usc.uscsi_timeout = 100;
#endif
	if (ioctl (fd, USCSICMD, &usc) == FAILURE)
		stat = FAILURE;
	else if ((SCSIStatus = usc.uscsi_status) == STATUS_GOOD)
	{
		SCSISenseKey   = buf [ 2] & 0x0F;
		SCSISenseCode  = buf [12];
		SCSISenseCodeQ = buf [13];
		stat = SUCCESS;
	}
	else
	{
#if 0
		if (SCSIStatus == STATUS_CHECK)
			RequestSense (fd, buf, buflen);
#endif
		stat = FAILURE;
	}
	return stat;
}

static int	SCSIPerform	(int fd, struct uscsi_cmd *pusc)

{
	int		stat;
	BYTE	sense_data [255];
	
#ifdef sunos5
	pusc->uscsi_flags |= USCSI_RQENABLE;
	pusc->uscsi_timeout = 100;
	pusc->uscsi_rqbuf = sense_data;
	pusc->uscsi_rqlen = sizeof sense_data;
#endif
	if (ioctl (fd, USCSICMD, pusc) == FAILURE)
		stat = FAILURE;
	else if ((SCSIStatus = pusc->uscsi_status) == STATUS_GOOD)
		stat = SUCCESS;
	else
	{
		stat = FAILURE;
#if 0
		if (SCSIStatus == STATUS_CHECK)
		{
			RequestSense (fd, sense_data, sizeof sense_data);
			SCSISenseKey   = sense_data [ 2] & 0x0F;
			SCSISenseCode  = sense_data [12];
			SCSISenseCodeQ = sense_data [13];
			if (SCSISenseKey == KEY_UNIT_ATTENTION && UnitAtnRetry)
				stat = SCSIPerform (fd, pusc);
		}
#endif
	}
	return stat;
}

int		_ReadWrite10	(int fd, DWORD lba, WORD count, void *buf, int func)

{
	static struct uscsi_cmd	usc;
	BYTE	cdb [10];

	memset (cdb, 0, sizeof cdb);
	memset (&usc, 0, sizeof usc);
	switch (func)
	{
		case 0:	cdb [0] = SCMD_READ_G1;			break;
		case 1:	cdb [0] = SCMD_WRITE_G1;		break;
		case 2:	cdb [0] = SCMD_WRITE_VERIFY;	break;
		case 3:	cdb [0] = 0x41;					break;
	}
	cdb [2] = lba >> 24;
	cdb [3] = lba >> 16;
	cdb [4] = lba >>  8;
	cdb [5] = lba;
	cdb [7] = count >> 8;
	cdb [8] = count;
	usc.uscsi_cdb = cdb;
	usc.uscsi_cdblen = sizeof cdb;
	usc.uscsi_bufaddr = buf;
	usc.uscsi_buflen = blocksize * (func == 3 ? 1 : count);
	usc.uscsi_flags = USCSI_DIAGNOSE | USCSI_ISOLATE | (func ? 0 : USCSI_READ);
	return SCSIPerform (fd, &usc);
}

int		TestUnitReady	(int fd)

{
	static struct uscsi_cmd	usc;
	static BYTE	cdb [6];

	memset (cdb, 0, sizeof cdb);
	memset (&usc, 0, sizeof usc);
	cdb [0] = SCMD_TEST_UNIT_READY;
	usc.uscsi_cdb = cdb;
	usc.uscsi_cdblen = sizeof cdb;
	usc.uscsi_bufaddr = NULL;
	usc.uscsi_buflen = 0;
	usc.uscsi_flags = USCSI_DIAGNOSE | USCSI_ISOLATE;
	return SCSIPerform (fd, &usc);
}

int		StartStopUnit	(int fd, int flag)

{
	static struct uscsi_cmd	usc;
	static BYTE	cdb [6];

	memset (cdb, 0, sizeof cdb);
	memset (&usc, 0, sizeof usc);
	cdb [0] = SCMD_START_STOP;
	cdb [4] = flag;
	usc.uscsi_cdb = cdb;
	usc.uscsi_cdblen = sizeof cdb;
	usc.uscsi_bufaddr = NULL;
	usc.uscsi_buflen = 0;
	usc.uscsi_flags = USCSI_DIAGNOSE | USCSI_ISOLATE;
	return SCSIPerform (fd, &usc);
}

int		PreventAllowMediumRemoval	(int fd, int flag)

{
	static struct uscsi_cmd	usc;
	static BYTE	cdb [6];

	memset (cdb, 0, sizeof cdb);
	memset (&usc, 0, sizeof usc);
	cdb [0] = SCMD_DOORLOCK;
	cdb [4] = flag;
	usc.uscsi_cdb = cdb;
	usc.uscsi_cdblen = sizeof cdb;
	usc.uscsi_bufaddr = NULL;
	usc.uscsi_buflen = 0;
	usc.uscsi_flags = USCSI_DIAGNOSE | USCSI_ISOLATE;
	return SCSIPerform (fd, &usc);
}

int		ModeSense		(int fd, BYTE *buf, int buflen)

{
	static struct uscsi_cmd	usc;
	static BYTE	cdb [6];
	
	if  (buflen > 255) buflen = 255;
	memset (cdb, 0, sizeof cdb);
	memset (&usc, 0, sizeof usc);
	cdb [0] = SCMD_MODE_SENSE;
	cdb [2] = 0x3F;
	cdb [4] = buflen;
	usc.uscsi_cdb = cdb;
	usc.uscsi_cdblen = sizeof cdb;
	usc.uscsi_bufaddr = buf;
	usc.uscsi_buflen = buflen;
	usc.uscsi_flags = USCSI_DIAGNOSE | USCSI_ISOLATE | USCSI_READ;
	return SCSIPerform (fd, &usc);
}

int		Inquiry	(int fd, BYTE *buf, int buflen)

{
	static struct uscsi_cmd	usc;
	static BYTE	cdb [6];
	
	if (buflen > 255) buflen = 255;
	memset (cdb, 0, sizeof cdb);
	memset (&usc, 0, sizeof usc);
	cdb [0] = SCMD_INQUIRY;
	cdb [4] = buflen;
	usc.uscsi_cdb = cdb;
	usc.uscsi_cdblen = sizeof cdb;
	usc.uscsi_bufaddr = buf;
	usc.uscsi_buflen = buflen;
	usc.uscsi_flags = USCSI_DIAGNOSE | USCSI_ISOLATE | USCSI_READ;
	return SCSIPerform (fd, &usc);
}

int		ReadCapacity	(int fd, DWORD *buf)

{
	static struct uscsi_cmd	usc;
	static BYTE	cdb [10];
	int	result;
	
	memset (cdb, 0, sizeof cdb);
	memset (&usc, 0, sizeof usc);
	cdb [0] = SCMD_READ_CAPACITY;
	usc.uscsi_cdb = cdb;
	usc.uscsi_cdblen = sizeof cdb;
	usc.uscsi_bufaddr = (caddr_t) buf;
	usc.uscsi_buflen = 8;
	usc.uscsi_flags = USCSI_DIAGNOSE | USCSI_ISOLATE | USCSI_READ;
	if ((result = SCSIPerform (fd, &usc)) == SUCCESS)
	{
		buf [0] = ntohl (buf [0]);
		buf [1] = ntohl (buf [1]);
	}
	return result;
}

int		FormatUnit		(int fd)

{
	static struct uscsi_cmd	usc;
	static BYTE	cdb [6];
	
	memset (cdb, 0, sizeof cdb);
	memset (&usc, 0, sizeof usc);
	cdb [0] = SCMD_FORMAT;
	usc.uscsi_cdb = cdb;
	usc.uscsi_cdblen = sizeof cdb;
	usc.uscsi_bufaddr = NULL;
	usc.uscsi_buflen = 0;
	usc.uscsi_flags = USCSI_DIAGNOSE | USCSI_ISOLATE;
	return SCSIPerform (fd, &usc);
}

