# common makefile for UNIX

MEI_RELEASE = 1.53
MOE_RELEASE = 1.53

MEI_DEFINES = $(SWITCHES) -DRELEASE=\"$(MEI_RELEASE)\" \
	-DDEVICE_FILE=\"$(DEVICE_FILE)\" \
	-DDEVICE_FILE_TMPL=\"$(DEVICE_FILE_TMPL)\" \
	-DLIBDIR=\"$(LIBDIR)\"
MOE_DEFINES = $(SWITCHES) -DRELEASE=\"$(MOE_RELEASE)\" \
	-DDEVICE_FILE=\"$(DEVICE_FILE)\" \
	-DDEVICE_FILE_TMPL=\"$(DEVICE_FILE_TMPL)\"
MEI_MAN_DEFINES = -D$(OSTYPE) -DRELEASE=$(MEI_RELEASE) \
	-DDEVICE_FILE=$(DEVICE_FILE) -DLIBDIR=$(LIBDIR)
MOE_MAN_DEFINES = -D$(OSTYPE) -DRELEASE=$(MOE_RELEASE) \
	-DDEVICE_FILE=$(DEVICE_FILE)

all:	mei moe readme.txt msdos.txt unix.txt

mei:	mei.o $(SCSI_MODULE).o $(SUPPLOBJS)
	$(CC) $(CFLAGS) -o mei mei.o $(SCSI_MODULE).o $(SUPPLOBJS)

moe:	moe.o $(SCSI_MODULE).o
	$(CC) $(CFLAGS) -o moe moe.o $(SCSI_MODULE).o

mei.o:	mei.c m_mei.h typedef.h $(SCSI_MODULE).h
	$(CC) $(CFLAGS) $(MEI_DEFINES) -c mei.c

moe.o:	moe.c m_moe.h typedef.h $(SCSI_MODULE).h
	$(CC) $(CFLAGS) $(MOE_DEFINES) -c moe.c

$(SCSI_MODULE).o:	$(SCSI_MODULE).c $(SCSI_MODULE).h typedef.h
	$(CC) $(CFLAGS) $(SWITCHES) -c $(SCSI_MODULE).c

perror.o:	perror.c typedef.h errno.h
	$(CC) $(CFLAGS) $(SWITCHES) -c perror.c

m_mei.h:	_en/mei.msg
	sed -f msg.sed _en/mei.msg > m_mei.h

m_moe.h:	_en/moe.msg
	sed -f msg.sed _en/moe.msg > m_moe.h

install:	all
	$(INSTALL) mei $(BINDIR)/mei
	$(INSTALL) moe $(BINDIR)/moe
	strip $(BINDIR)/mei
	strip $(BINDIR)/moe

install.man:	$(MAN_TARGETS)

install.en:	$(LN_EN)/mei.$(MANEXT) $(LN_EN)/moe.$(MANEXT)
	$(RM) $(MANDIR)/cat$(MANEXT)/mei.$(MANEXT)
	$(RM) $(MANDIR)/cat$(MANEXT)/moe.$(MANEXT)
	$(INSTALL) $(LN_EN)/mei.$(MANEXT) $(MANDIR)/man$(MANEXT)
	$(INSTALL) $(LN_EN)/moe.$(MANEXT) $(MANDIR)/man$(MANEXT)

install.ja:	$(LN_EUCJP)/mei.$(MANEXT) $(LN_EUCJP)/moe.$(MANEXT)
	$(RM) $(MANDIR)/$(LN_EUCJP)/cat$(MANEXT)/mei.$(MANEXT)
	$(RM) $(MANDIR)/$(LN_EUCJP)/cat$(MANEXT)/moe.$(MANEXT)
	$(INSTALL) $(LN_EUCJP)/mei.$(MANEXT) $(MANDIR)/$(LN_EUCJP)/man$(MANEXT)
	$(INSTALL) $(LN_EUCJP)/moe.$(MANEXT) $(MANDIR)/$(LN_EUCJP)/man$(MANEXT)

$(LN_EN)/mei.$(MANEXT):	_en/mei.src
	-mkdir $(LN_EN)
	LANG=$(LN_EN); export LANG;\
	cat _en/mei.src |\
	$(CPP) $(MEI_MAN_DEFINES) -DDATE="`date +%x`" |\
	egrep -v '^$$' > $(LN_EN)/mei.$(MANEXT)

$(LN_EUCJP)/mei.$(MANEXT):	_ja/mei.src
	-mkdir $(LN_EUCJP)
	LANG=$(LN_EUCJP); export LANG;\
	$(KCC) -e _ja/mei.src |\
	$(CPP) $(MEI_MAN_DEFINES) -DDATE="`date +%x`" |\
	egrep -v '^$$' > $(LN_EUCJP)/mei.$(MANEXT)

$(LN_EN)/moe.$(MANEXT):	_en/moe.src
	-mkdir $(LN_EN)
	LANG=$(LN_EN); export LANG;\
	cat _en/moe.src |\
	$(CPP) $(MOE_MAN_DEFINES) -DDATE="`date +%x`" |\
	egrep -v '^$$' > $(LN_EN)/moe.$(MANEXT)

$(LN_EUCJP)/moe.$(MANEXT):	_ja/moe.src
	-mkdir $(LN_EUCJP)
	LANG=$(LN_EUCJP); export LANG;\
	$(KCC) -e _ja/moe.src |\
	$(CPP) $(MOE_MAN_DEFINES) -DDATE="`date +%x`" |\
	egrep -v '^$$' > $(LN_EUCJP)/moe.$(MANEXT)

install.cat:
	-mkdir $(LN_EN)
	$(CP) _en/mei.msg $(LN_EN)
	$(CP) _en/moe.msg $(LN_EN)
	-if [ /$(LN_SJIS) != / ]; then\
		mkdir $(LN_SJIS);\
		$(KCC) -s _ja/mei.msg > $(LN_SJIS)/mei.msg;\
		$(KCC) -s _ja/moe.msg > $(LN_SJIS)/moe.msg;\
	fi
	-if [ /$(LN_EUCJP) != / ]; then\
		mkdir $(LN_EUCJP);\
		$(KCC) -e _ja/mei.msg > $(LN_EUCJP)/mei.msg;\
		$(KCC) -e _ja/moe.msg > $(LN_EUCJP)/moe.msg;\
	fi
	-if [ /$(LN_FR) != / ]; then\
		mkdir $(LN_FR);\
		LANG=$(LN_FR); export LANG;\
		sed -f 850_ltn1.sed _fr/mei.msg > $(LN_FR)/mei.msg;\
		sed -f 850_ltn1.sed _fr/moe.msg > $(LN_FR)/moe.msg;\
	fi
	for l in $(CAT_LOCALES); do\
		cd $$l;\
		catpath=`echo $(NLSDIR)/|sed s/%L/$$l/|tr . $(LOCDIR_DELIM)`;\
		$(RM) mei.cat;\
		$(RM) moe.cat;\
		LANG=$$l;\
		export LANG;\
		gencat mei.cat mei.msg;\
		gencat moe.cat moe.msg;\
		$(INSTALL) mei.cat $$catpath;\
		$(INSTALL) moe.cat $$catpath;\
		cd ..;\
	done

readme.txt:	readme.ntf
	ntf -t -o readme.txt readme.ntf

msdos.txt:	msdos.ntf
	ntf -t -o msdos.txt msdos.ntf

unix.txt:	unix.ntf
	ntf -t -o unix.txt unix.ntf

clean:
	$(RM) *.o core \
		mei mei.man m_mei.h \
		moe moe.man m_moe.h
	$(RM) -r $(CAT_LOCALES)

distrib:	distrib.doc distrib.src

distrib.doc:
	PAGER=cat man $(MANEXT) mei | col -b | $(KCC) -s > mei.man
	PAGER=cat man $(MANEXT) moe | col -b | $(KCC) -s > moe.man
	$(RM) mei_d.tgz
	tar cf mei_d.tar *.txt *.man instcat.bat _en _ja _fr
	gzip mei_d.tar
	mv mei_d.tar.gz mei_d.tgz

distrib.src:
	$(RM) mei_s.tgz
	tar cf mei_s.tar *.mak *.c *.h *.def *.sed *.ntf
	gzip mei_s.tar
	mv mei_s.tar.gz mei_s.tgz

# FreeBSD packages
package:
	$(RM) mei_pkg.tgz
	tar cf mei_pkg.tar \
		+* \
		-C /usr/local \
		bin/mei bin/moe \
		man/man1/mei.1 man/man1/moe.1 \
		man/$(LN_EUCJP)/man1/mei.1 man/$(LN_EUCJP)/man1/moe.1 \
		-C /usr/share \
		nls/$(LN_EN)/mei.cat nls/$(LN_EN)/moe.cat \
		nls/$(LN_EUCJP)/mei.cat nls/$(LN_EUCJP)/moe.cat
	gzip mei_pkg.tar
	mv mei_pkg.tar.gz mei_pkg.tgz
